# pyOCD debugger
# Copyright (c) 2020 Arm Limited
# Copyright (c) 2021-2023 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from contextlib import contextmanager
import lark.lark
import lark.exceptions
import lark.visitors
import logging
import threading
from dataclasses import dataclass
from enum import Enum
from inspect import signature
from lark.lexer import Token as LarkToken
from lark.tree import Tree as LarkTree
from typing import (Any, Iterator, cast, List, Optional, Union, TYPE_CHECKING)
from typing_extensions import Self

from ...core import exceptions
from ...coresight.ap import (APv1Address, APv2Address)
from ...utility.graph import GraphNode
from ...utility.mask import bit_invert
from ...utility.timeout import Timeout
from .scope import Scope

if TYPE_CHECKING:
    # Only import Session when type checking to avoid complex import cycles.
    from ...core.session import Session
    from ...coresight.ap import APAddressBase
    from .delegates import DebugSequenceDelegate

LOG = logging.getLogger(__name__)
TRACE = LOG.getChild("trace")
TRACE.setLevel(logging.CRITICAL)

NodeType = Union[LarkTree, LarkToken, int]

class DebugSequenceError(exceptions.Error):
    pass

class DebugSequenceSemanticError(DebugSequenceError):
    pass

class DebugSequenceRuntimeError(exceptions.Error):
    pass

def _is_token(tok: Any, typename: str) -> bool:
    """@brief Test whether a node is a specific type of token."""
    return isinstance(tok, LarkToken) and tok.type == typename

class _ConvertLiterals(lark.visitors.Transformer):
    """@brief Transformer to convert integer literal tokens to integers.

    Running this transformer during the parse is more efficient than handling it post-parse
    such as during optimization.
    """
    def INTLIT(self, tok: LarkToken) -> int: # pylint: disable=invalid-name
        return int(tok.value.upper().rstrip('U'), base=0)

    def STRLIT(self, tok: LarkToken) -> LarkToken:
        tok.value = tok.value.strip('"')
        return tok

class Parser:
    """@brief Debug sequence statement parser."""

    ## Shared parser object.
    _parser = lark.lark.Lark.open("sequences.lark",
                        rel_to=__file__,
                        parser="lalr",
                        maybe_placeholders=True,
                        propagate_positions=True,
                        transformer=_ConvertLiterals())

    @classmethod
    def parse(cls, data: str) -> LarkTree:
        try:
            # Parse the input.
            tree = cls._parser.parse(data)

            # Return the resulting tree.
            return tree
        except lark.exceptions.UnexpectedInput as e:
            message = str(e) + "\n\nContext: " + e.get_context(data, 40)
            raise exceptions.Error(message) from e

class DebugSequenceExecutionContext:
    """@brief Context for running debug sequences.

    Instances of this class contain state for running a call stack of sequences. There is a stack
    of sequence node and scope pairs. This stack is also used to get the current running sequence.

    Context objects can be used as context managers, which will set the context as the current
    thread's active context. The class method get_active_context() returns the currently active
    context for the current thread. There is no stack of active contexts; the assumption is that
    only one call stack of debug sequences can be run at any time on a given thread.
    """

    _thread_local_contexts = threading.local()

    @dataclass
    class _NodeScopeStackItem:
        node: DebugSequenceNode
        scope: Scope

    @classmethod
    def get_active_context(cls) -> DebugSequenceExecutionContext:
        """@brief Return the active context for the current thread.
        @exception AttributeError There is no active context for this thread.
        """
        return cls._thread_local_contexts.context

    def __init__(self, session: Session, delegate: DebugSequenceDelegate, pname: Optional[str]) -> None:
        """@brief Constructor.

        @param self
        @param session Session for the target connection.
        @param delegate Debug sequence delegate that owns the running sequences.
        @param pname Pname under which sequences are being run.
        """
        self._session = session
        self._delegate = delegate
        self._default_ap_address = APv1Address(0)
        self._pname = pname
        self._stack: List[DebugSequenceExecutionContext._NodeScopeStackItem] = []

    @property
    def session(self) -> Session:
        return self._session

    @property
    def delegate(self) -> DebugSequenceDelegate:
        return self._delegate

    @property
    def default_ap(self) -> APAddressBase:
        """@brief Return the AP address from which the __ap and __apid variables should be inited.

        The initial AP address is the one defined for the <debug> element corresponding to the active
        processor.

        This is contained in the context rather than returned by the sequence delegate because it can
        differ for each execution of a debug sequence. Compare with the delegate methods for getting
        `__protocol`, `__connection`, and `__traceout`, which should be static for an entire session.
        """
        return self._default_ap_address

    @default_ap.setter
    def default_ap(self, address: APAddressBase) -> None:
        """@brief Set the default AP adddress."""
        self._default_ap_address = address

    @property
    def pname(self) -> Optional[str]:
        """@brief Pname for which sequences are executing in this context.

        The Pname for this context is set when the context is created.
        Only debug sequences with this Pname or an empty Pname can be run in this context.
        """
        return self._pname

    @property
    def current_scope(self) -> Scope:
        """@brief Topmost scope of the scope stack.
        @exception AssertionError Raised if an attempt is made to get the current scope without a
            currently running debug sequence. If needed, this can be checked before accessing this
            property by getting the `.has_current_sequence` property.
        """
        assert self._stack, "current_scope accessed from outside a running sequence"
        return self._stack[-1].scope

    @property
    def has_current_sequence(self) -> bool:
        """@brief Whether there is a debug sequence running on this context.

        By definition, there must be a debug sequence on the stack if the stack is not empty,
        with one exception for when the debugvars block is run to create the the root scope.
        """
        return bool(self._stack)

    @property
    def current_sequence(self) -> DebugSequence:
        """@brief Currently executing debug sequence."""
        assert self._stack, "current_sequence accessed without a running debug sequence"

        # Walk the stack to find the most recent DebugSequence.
        for elem in reversed(self._stack):
            if isinstance(elem.node, DebugSequence):
                return elem.node

        # Ooh, not good..
        assert False, "invalid state: no debug sequence on active debug sequence execute context stack"

    def _push(self, node: DebugSequenceNode, scope: Scope) -> None:
        """@brief Context stack push operation.

        Separated even though it's tiny so it can be used manually for unit testing.
        """
        self._stack.append(self._NodeScopeStackItem(node, scope))

    def _pop(self) -> None:
        """@brief Context stack pop operation.

        Separated even though it's tiny so it can be used manually for unit testing.
        """
        self._stack.pop()

    @contextmanager
    def push(self, node: DebugSequenceNode, scope: Scope) -> Iterator[None]:
        """@brief Context manager to push/pop a sequence node and scope pair on the context stack.

        A node/scope pair is pushed prior to that node (and its children) being executed.

        Scopes in the stack do not automatically or necessarily have parent links between them. This
        allows for unlinked scopes from the call stack of sequences to be placed on the stack, without
        causing issues with variable access.

        Blocks are not normally pushed, since they do not define scopes. The sole exception to this is when
        the debugvars block is executed to create the root scope, but this is done by the code that
        creates the root scope and not the Block itself.

        When the context manager exits, the scope is popped.
        """
        try:
            self._push(node, scope)
            yield
        finally:
            self._pop()

    def get_variable(self, name: str) -> int:
        """@brief Return the value of a variable from the current scope."""
        assert self.current_scope is not None
        try:
            return self.current_scope.get(name)
        except KeyError as err:
            LOG.debug("debug sequence reference to undefined variable %s... %s", name, self.current_scope.dump())
            raise DebugSequenceRuntimeError(f"reference to undefined variable {name}") from err

    def __enter__(self) -> Self:
        """@brief Make this context the active one for the current thread."""
        self._thread_local_contexts.context = self
        return self

    def __exit__(self, exc_type, value, traceback) -> None:
        """@brief Clear the current thread's active context."""
        assert self._thread_local_contexts.context == self
        del self._thread_local_contexts.context

class DebugSequenceNode(GraphNode):
    """@brief Common base class for debug sequence nodes."""

    def __init__(self, info: str = "") -> None:
        super().__init__()
        self._info = info

    @property
    def info(self) -> str:
        return self._info

    def execute(self, context: DebugSequenceExecutionContext) -> Optional[Scope]:
        """@brief Execute the sequence node.

        @return If a scope was created in order to execute the sequence node, that scope is returned
            to the caller so any variables can be accessed.
        """
        raise NotImplementedError()

    def _execute_children(self, context: DebugSequenceExecutionContext) -> None:
        """@brief Execute all child nodes."""
        for node in self.children:
            cast(DebugSequenceNode, node).execute(context)

class DebugSequence(DebugSequenceNode):
    """@brief Named debug sequence.

    Variable scoping:
    - Sequences and control elements create new scopes.
    - Scope extends to child control elements.
    - Block elements do not create a new scope.
    - Variables in a parent scope can be modified.
    - Leaving a scope destroys contained variables.
    - Variables are not passed to sub-sequences.

    Special read-write variables:
    - __dp, __ap, __apid, __errorcontrol
        - Propagates to another sequence called via the Sequence() function.
        - Restored to previous values when sub-sequence returns.
    - __Result
        - Not pushed when calling another sequence.
        - 0=success

    Special read-only variables:
    - __protocol
        - [15:0] 0=error, 1=JTAG, 2=SWD, 3=cJTAG
        - [16] SWJ-DP present?
        - [17] switch through dormant state?
    - __connection
        - [7:0] connection type: 0=error/disconnected, 1=for debug, 2=for flashing
        - [15:8] reset type: 0=error, 1=hw, 2=SYSRESETREQ, 3=VECTRESET
        - [16] connect under reset?
        - [17] pre-connect reset?
    - __traceout
        - [0] SWO enabled?
        - [1] parallel trace enabled?
        - [2] trace buffer enabled?
        - [21:16] selected parallel trace port size
    - __FlashOp
        - 0=no op, 1=erase full chip, 2=erase sector, 3=program
    - __FlashAddr
    - __FlashLen
    - __FlashArg
    """

    ## Special predefined variables that are read-write and propagate to sub-sequences.
    _SPECIAL_VARS = ['__dp', '__ap', '__apid', '__errorcontrol']

    ## Predefined variables that are read-write, but do not propagate.
    _WRITABLE_PREDEFINED = ['__Result']

    def __init__(
            self,
            name: str,
            is_enabled: bool = True,
            pname: Optional[str] = None,
            info: str = ""
            ) -> None:
        super().__init__(info)
        self._name = name
        self._is_enabled = is_enabled
        self._pname = pname

    @property
    def name(self) -> str:
        return self._name

    @property
    def pname(self) -> Optional[str]:
        return self._pname

    @property
    def is_enabled(self) -> bool:
        return self._is_enabled

    def _create_scope(self, context: DebugSequenceExecutionContext) -> Scope:
        """@brief Create a new variables scope with predefined variables filled in."""
        delegate = context.delegate

        # Create the new scope using the delegate's root scope as parent.
        scope = Scope(parent=context.delegate.get_root_scope(context), name=self.name)

        # Writable result.
        scope.set('__Result', 0)

        # Propagate specials if there is a sequence that called us.
        if context.has_current_sequence:
            scope.copy_variables(context.current_scope, self._SPECIAL_VARS)
        # Otherwise just fill in the defaults.
        else:
            scope.set('__errorcontrol', 0)

            # Convert the default AP address to __ap and __apid variables. If the AP is v1 then
            # __ap is set, for v2 __apid is set. The other variable gets set to 0.
            default_ap_address = context.default_ap
            scope.set('__dp', default_ap_address.dp_index) # We still only support one DP.
            scope.set('__ap', default_ap_address.nominal_address
                                if isinstance(default_ap_address, APv1Address)
                                else 0)
            scope.set('__apid',  default_ap_address.nominal_address
                                    if isinstance(default_ap_address, APv2Address)
                                    else 0)

        # Generate __protocol value.
        protocol = delegate.get_protocol()
        scope.set('__protocol', protocol, readonly=True)

        # Generate __connection value.
        connection = delegate.get_connection_type()
        scope.set('__connection', connection, readonly=True)

        # Generate __traceout value.
        traceout = delegate.get_traceout()
        scope.set('__traceout', traceout, readonly=True)

        # Flash algorithm sequence parameters.
        scope.set('__FlashOp', 0, readonly=True)
        scope.set('__FlashAddr', 0, readonly=True)
        scope.set('__FlashLen', 0, readonly=True)
        scope.set('__FlashArg', 0, readonly=True)
        return scope

    def execute(self, context: DebugSequenceExecutionContext) -> Optional[Scope]:
        """@brief Run the sequence."""
        scope = self._create_scope(context)

        # Make this the active sequence.
        with context.push(self, scope):
            self._execute_children(context)

        return scope

    def __eq__(self, o: object) -> bool:
        return (isinstance(o, DebugSequence)
                and self.name == o.name
                and self.pname == o.pname
                and self.is_enabled == o.is_enabled)

    def __hash__(self) -> int:
        return hash((self.name, self.pname, self.is_enabled))

    def __repr__(self):
        return f"<{type(self).__name__}@{id(self):x} {self.name} enabled={self.is_enabled} pname={self.pname}>"

class Control(DebugSequenceNode):
    """@brief Base class for control nodes of debug sequences.

    Control elements create new scopes.
    """

    class ControlType(Enum):
        IF = 1
        WHILE = 2

    def __init__(self, control_type: ControlType, predicate: str, info: str = "", timeout_µs: int = 0) -> None:
        """@brief Constructor.
        @param self The control object.
        @param control_type One of the #ControlType enums that selects between if- and while-type.
        @param predicate String of the predicate expression.
        @param info Optional descriptive string.
        @param timeout_µs Integer timeout in microseconds. A value of zero means an infinite timeout.
        """
        super().__init__(info)
        self._type = control_type
        # Convert µs to seconds, and 0 to None.
        self._timeout = (timeout_µs / 1000000) if timeout_µs else None
        self._predicate = predicate
        self._ast = Parser.parse(predicate)

    def execute(self, context: DebugSequenceExecutionContext) -> Optional[Scope]:
        """@brief Run the sequence."""
        # Get our scope and interpreter objects.
        parent_scope = context.current_scope
        scope = Scope(
            parent_scope,
            name=f"{parent_scope.name}.{self._type.name}"
            )
        interp = Interpreter(self._ast, scope, context)

        # Push our new scope.
        with context.push(self, scope):
            # Start the timeout counting.
            timeout = Timeout(self._timeout)
            timeout.start()

            # Execute the predicate a first time.
            result = interp.execute()
            TRACE.debug("%s(%s): pred=%s", self._type.name, self._predicate, result)

            while result and timeout.check():
                # Execute all child nodes.
                self._execute_children(context)

                # For an if control, we're done.
                if self._type == self.ControlType.IF:
                    break
                # For a while control, re-evaluate the predicate.
                elif self._type == self.ControlType.WHILE:
                    result = interp.execute()
                    TRACE.debug("%s(%s): pred=%d", self._type.name, self._predicate, result)

        return scope

    def __repr__(self):
        return f"<{type(self).__name__}@{id(self):x} {self._ast.pretty()}>"

class WhileControl(Control):
    """@brief Looping debug sequence node."""

    def __init__(self, predicate: str, info: str = "", timeout: int = 0) -> None:
        super().__init__(self.ControlType.WHILE, predicate, info, timeout)

class IfControl(Control):
    """@brief Conditional debug sequence node."""

    def __init__(self, predicate: str, info: str = "", timeout: int = 0) -> None:
        super().__init__(self.ControlType.IF, predicate, info, timeout)

class Block(DebugSequenceNode):
    """@brief Block of debug sequence statements.

    Block elements do not create a new scope.
    """

    def __init__(self, code: str, is_atomic: bool = False, info: str = "") -> None:
        super().__init__(info)
        self._ast = Parser.parse(code)
        self._is_atomic = is_atomic

    def execute(self, context: DebugSequenceExecutionContext) -> Optional[Scope]:
        """@brief Run the sequence."""
        assert context.session.probe

        try:
            # If the block is atomic, hold the probe lock while it is executed.
            if self._is_atomic:
                context.session.probe.lock()

            interp = Interpreter(self._ast, context.current_scope, context)
            interp.execute()
        finally:
            if self._is_atomic:
                context.session.probe.unlock()

    def __repr__(self):
        atomic_str = " atomic" if self._is_atomic else ""
        return f"<{type(self).__name__}@{id(self):x}{atomic_str} {self._ast.pretty()}>"

# Using Any type for the methods of this class is a workaround for LarkToken not being
# handled or inferred correctly, since Lark doesn't have annotations.
class _ConstantFolder(lark.visitors.Transformer):
    """@brief Performs basic constant folding on expressions."""

    def _is_intlit(self, node: Any) -> bool:
        return isinstance(node, int)

    def ternary_expr(self, children: Any) -> Any:
        predicate = children[0]
        true_expr = children[1]
        false_expr = children[2]

        # Fold ternaries to either true or false branch when the predicate is an integer constant.
        if self._is_intlit(predicate):
            if predicate != 0:
                return true_expr
            elif predicate == 0:
                return false_expr

        return LarkTree('ternary_expr', children)

    def binary_expr(self, children: Any) -> Any:
        left = children[0]
        op = children[1].value
        right = children[2]

        # Fold binary expressions on literals.
        if self._is_intlit(left) and self._is_intlit(right):
            result = _BINARY_OPS[op](left, right)
            # TRACE.debug("opt: %#x %s %#x -> %#x", left, op, right, result)
            return result

        # Fold binary expressions with a left operand of zero.
        elif self._is_intlit(right) and right == 0:
            # Operators whose result will be the left operand unmodified.
            if op in ('+', '-', '|', '^', '<<', '>>', '||'):
                # TRACE.debug("opt: x %s 0 -> x", op)
                return left
            # Operators whose result will be zero.
            elif op in ('*', '/', '%', '&', '&&'):
                # TRACE.debug("opt: x %s 0 -> 0", op)
                return 0

        # Fold binary expression with a right operand of zero.
        elif self._is_intlit(left) and left == 0:
            # Operators whose result will be the right operand unmodified.
            if op in ('+', '-', '|', '^', '||'):
                # TRACE.debug("opt: 0 %s x -> x", op)
                return right
            # Operators whose result will be zero.
            elif op in ('*', '/', '%', '&', '<<', '>>', '&&'):
                # TRACE.debug("opt: 0 %s x -> 0", op)
                return 0

        # Fold binary expressions with a left operand of 1.
        elif self._is_intlit(right) and right == 1:
            # Operators whose result will be the left operand unmodified.
            if op in ('*', '/'):
                # TRACE.debug("opt: x %s 1 -> x", op)
                return left
            # Operators whose result will be one.
            elif op in ('||',):
                # TRACE.debug("opt: x %s 1 -> 1", op)
                return 1
            # Operators whose result will be zero.
            elif op in ('%',):
                # TRACE.debug("opt: x %s 1 -> 0", op)
                return 0

        return LarkTree('binary_expr', children)

    def unary_expr(self, children: Any) -> Any:
        op = children[0].value
        arg = children[1]

        # Fold unary expressions on a literal.
        if self._is_intlit(arg):
            result = _UNARY_OPS[op](arg)
            # TRACE.debug("opt: %s %#x -> %#x", op, arg, result)
            return result

        return LarkTree('unary_expr', children)

## Lambdas for evaluating binary operators.
#
# Note that divide and modulo by 0 just results in 0 rather than an exception.
_BINARY_OPS = {
    '+':    lambda l, r: l + r,
    '-':    lambda l, r: l - r,
    '*':    lambda l, r: l * r,
    '/':    lambda l, r: 0 if (r == 0) else (l // r),
    '%':    lambda l, r: 0 if (r == 0) else (l % r),
    '&':    lambda l, r: l & r,
    '|':    lambda l, r: l | r,
    '^':    lambda l, r: l ^ r,
    '<<':   lambda l, r: l << r,
    '>>':   lambda l, r: l >> r,
    '&&':   lambda l, r: int(bool(l) & bool(r)), # implement C-style AND
    '||':   lambda l, r: int(bool(l) | bool(r)), # implement C-style OR
    '==':   lambda l, r: int(l == r),
    '!=':   lambda l, r: int(l != r),
    '>':    lambda l, r: int(l > r),
    '>=':   lambda l, r: int(l >= r),
    '<':    lambda l, r: int(l < r),
    '<=':   lambda l, r: int(l <= r),
    }

## Lambdas for evaluating unary operators.
_UNARY_OPS = {
    '~':    lambda v: bit_invert(v, width=64),
    '!':    lambda v: int(not v),
    '+':    lambda v: v,
    '-':    lambda v: (-v) & 0xffffffffffffffff, # Mask to get unsigned two's complement.
    }

class SemanticChecker:
    """@brief Check the semantics of debug sequence statements."""

    class _SemanticsVisitor(lark.visitors.Visitor):
        """@brief Visitor for performing semantic checks of debug sequence statements."""

        def __init__(self, scope: Scope, context: DebugSequenceExecutionContext) -> None:
            super().__init__()
            self._scope = scope
            self._context = context
            self._fns = self._context.delegate.get_sequence_functions()
            self._declared_variables = set()

        def decl_stmt(self, tree: LarkTree) -> None:
            # Record the declared variable name.
            assert _is_token(tree.children[0], 'IDENT')
            assert isinstance(tree.children[0], LarkToken)
            name = tree.children[0].value
            self._declared_variables.add(name)

            # Disallow assigning expressions consisting of only a string.
            if _is_token(tree.children[1], 'STRLIT'):
                raise DebugSequenceSemanticError(
                        f"line {tree.meta.line}: cannot store a string to variable '{name}'")

        def assign_expr(self, tree: LarkTree) -> None:
            # Assigned variable must have been previously declared.
            # TODO disabled until declarations are fully tracked in scopes.
            assert _is_token(tree.children[0], 'IDENT')
            assert isinstance(tree.children[0], LarkToken)
            name = tree.children[0].value
#             if name not in self._declared_variables:
#                 raise DebugSequenceSemanticError(
#                         f"line {tree.meta.line}: attempt to set undeclared variable '{name}'")

            # Disallow assigning expressions consisting of only a string.
            if _is_token(tree.children[2], 'STRLIT'):
                raise DebugSequenceSemanticError(
                        f"line {tree.meta.line}: cannot store a string to variable '{name}'")

        def expr_stmt(self, tree: LarkTree) -> None:
            # Disallow statements consisting of only a string.
            if _is_token(tree.children[0], 'STRLIT'):
                raise DebugSequenceSemanticError(
                        f"line {tree.meta.line}: expression statements consisting of only a string are invalid")

        def fncall(self, tree: LarkTree) -> None:
            fn_name = tree.children[0]
            assert isinstance(fn_name, str)
            # TRACE.debug("checking %s (%s)", fn_name, tree.children[1:])

            # Case-insensitive match.
            fn_name = fn_name.lower()

            # Look up the function on the delegate.
            try:
                impl = getattr(self._fns, fn_name)
            except AttributeError:
                raise DebugSequenceSemanticError(f"line {tree.meta.line}: call to unknown function '{fn_name}'")

            # Get the function's signature.
            sig = signature(impl)

            arg_count = len(tree.children[1:])
            param_count = 0
            has_varargs = False

            # Note that the 'self' parameter should not be present due to getting the signature from the
            # bound method of the functions delegate instance.
            for param in sig.parameters.values():
                if param.kind == param.VAR_POSITIONAL:
                    # Don't need to check past a varargs parameter.
                    has_varargs = True
                    break
                else:
                    param_count += 1

                # Check arg count.
                if param_count > arg_count:
                    raise DebugSequenceSemanticError(
                            f"line {tree.meta.line}: function '{fn_name}' is passed too few arguments")

                # Check argument types.
                #
                # The type of the argument is only checked if a type annotation is present. Type annotations
                # must be stringified, eg using 'from __future__ import annotations' or Python 3.10. Only
                # explicit 'str' and 'int' annotation types are currently supported; subclasses of these
                # types are not (which is why empty annotations are allowed, as a workaround).
                arg_node = tree.children[param_count]

                # str params require a literal string arg.
                if param.annotation == 'str' and not _is_token(arg_node, 'STRLIT'):
                    raise DebugSequenceSemanticError(
                            f"line {tree.meta.line}: function '{fn_name}' parameter '{param.name}' "
                            "requires a string argument")
                # int params require either a literal int or an expression tree.
                elif param.annotation == 'int' and not \
                        (isinstance(arg_node, int) or _is_token(arg_node, 'IDENT') or isinstance(arg_node, LarkTree)):
                    raise DebugSequenceSemanticError(
                            f"line {tree.meta.line}: function '{fn_name}' parameter '{param.name}' "
                            "requires an integer argument")

            # Check for more args than parameters.
            if (param_count < arg_count) and not has_varargs:
                raise DebugSequenceSemanticError(
                        f"line {tree.meta.line}: function '{fn_name}' is passed too many arguments")

            # Function-specific checks.
            if fn_name == "Sequence":
                # From above, we know the correct number of arguments is present, and that the arg is a string.
                assert isinstance(tree.children[1], LarkToken)
                name = tree.children[1].value

                # Look for a sequence with the given name.
                if not self._context.delegate.has_sequence_with_name(name, self._context.pname):
                    raise DebugSequenceSemanticError(
                            f"line {tree.meta.line}: attempt to call undefined sequence '{name}'")

    def __init__(self, tree: LarkTree, scope: Scope, context: DebugSequenceExecutionContext) -> None:
        """@brief Constructor.

        @param self This object.
        @param tree The abstract syntax tree that will be checked.
        @param scope Scope within which the AST will execute.
        @param delegate Delegate providing debug sequence function implementations.
        """
        super().__init__()
        self._tree = tree
        self._scope = scope
        self._context = context

    def check(self) -> None:
        """@brief Performs the semantic checks and raises for any errors.
        @exception DebugSequenceSemanticError A semantic error was discovered in the provided code.
        """
        visitor = self._SemanticsVisitor(self._scope, self._context)
        visitor.visit(self._tree)

class Interpreter:
    """@brief Interpreting for debug sequence ASTs.

    This class interprets the AST from only a single block or control node. The user of this class
    is required to handle crossing block/control boundaries.

    An Interpreter instance can be used to execute the AST more than once.
    """

    class _InterpreterVisitor(lark.visitors.Interpreter):
        """@brief Visitor for interpreting sequence trees."""

        def __init__(self, scope: Scope, context: DebugSequenceExecutionContext) -> None:
            super().__init__()
            self._scope = scope
            self._context = context
            self._fns = self._context.delegate.get_sequence_functions()

        def start(self, tree: LarkTree) -> Optional[int]:
            # Interpret the tree.
            values = self.visit_children(tree)
            return values.pop() if len(values) else None

        def _log_children(self, name: str, children: List) -> None: # pragma: no cover
            LOG.info('%s: %s', name, [(("Node: %s" % c.data) if hasattr(c, 'data') else ("%s=%s" % (c.type, c.value))) for c in children])

        def decl_stmt(self, tree: LarkTree) -> None:
            values = self.visit_children(tree)

            assert _is_token(values[0], 'IDENT')
            name = values[0].value
            # Handle __var declarations with no initialiser expression. Even though this is disallowed
            # by the specification, it appears in some DFPs, including some of NXP's.
            if values[1] is None:
                value = 0

                TRACE.debug("(line %d): decl %s = 0", getattr(tree.meta, 'line', 0), name)
            else:
                value = self._get_atom(values[1])

                TRACE.debug("(line %d): decl %s = %s", getattr(tree.meta, 'line', 0), name, self._format_atom(values[1]))

            self._scope.set(name, value)

        def assign_expr(self, tree: LarkTree) -> int:
            values = self.visit_children(tree)

            name = values[0].value
            op = values[1].value
            value = self._get_atom(values[2])

            TRACE.debug("(line %d): %s %s %s", getattr(tree.meta, 'line', 0), name, op, self._format_atom(values[2]))

            # Handle compound assignment operators.
            if op != '=':
                left = self._scope.get(name)
                op = op.rstrip('=')
                value = _BINARY_OPS[op](left, value)

            self._scope.set(name, value)

            # Return the variable's value as the assignment expression's value.
            return value

        def expr_stmt(self, tree: LarkTree) -> int:
            values = self.visit_children(tree)
            expr_value = values.pop()
            TRACE.debug("(line %d): expr stmt = %s", getattr(tree.meta, 'line', 0), self._format_atom(expr_value))
            return self._get_atom(expr_value)

        def ternary_expr(self, tree: LarkTree) -> int:
            values = self.visit_children(tree)

            predicate = self._get_atom(values[0])

            if not isinstance(predicate, int):
                raise DebugSequenceSemanticError("ternary expression predicate is not an integer")

            if predicate != 0:
                result = self._get_atom(values[1])
            else:
                result = self._get_atom(values[2])

            TRACE.debug("(line %s): %s ? %s : %s -> %s",
                    getattr(tree.meta, 'line', 0),
                    self._format_atom(values[0]),
                    self._format_atom(values[1]),
                    self._format_atom(values[2]),
                    hex(result))

            return result

        def binary_expr(self, tree: LarkTree) -> int:
            values = self.visit_children(tree)

            left = self._get_atom(values[0])
            op = values[1].value
            right = self._get_atom(values[2])

            result = _BINARY_OPS[op](left, right)

            TRACE.debug("(line %s): %s %s %s -> %s", getattr(tree.meta, 'line', 0),
                    self._format_atom(values[0]), op, self._format_atom(values[2]), hex(result))
            return result

        def unary_expr(self, tree: LarkTree) -> int:
            values = self.visit_children(tree)

            op = values[0].value
            value = self._get_atom(values[1])

            result = _UNARY_OPS[op](value)

            TRACE.debug("(line %s): %s %s -> %s", getattr(tree.meta, 'line', 0), op, self._format_atom(values[1]),
                hex(result))

            return result

        def fncall(self, tree: LarkTree) -> int:
            values = self.visit_children(tree)
            fn_name = values[0]
            fn_args = [self._get_atom(a) for a in values[1:]]

            # Case-insensitive match.
            fn_name = fn_name.lower()

            TRACE.debug("(line %d): fn %s (%s) ...", getattr(tree.meta, 'line', 0), fn_name,
                ", ".join(self._format_atom(a) for a in values[1:]))

            # Should have already verified the function name.
            impl = getattr(self._fns, fn_name)
            result = impl(*fn_args)
            if result is None:
                result = 0

            TRACE.debug("(line %d): fn %s () returned %s", getattr(tree.meta, 'line', 0), fn_name, hex(result))
            return result

        def _get_atom(self, node: NodeType) -> int:
            if isinstance(node, LarkTree):
                raise DebugSequenceSemanticError(f"expected atom but found an expression tree of type {node.data}")
            elif isinstance(node, LarkToken):
                if node.type == 'IDENT':
                    try:
                        return self._scope.get(node.value)
                    except KeyError as err:
                        LOG.debug("debug sequence reference to undefined variable %s... %s",
                                node.value, self._scope.dump())
                        raise DebugSequenceSemanticError(f"reference to undefined variable {node.value}") from err
                elif node.type in ('INTLIT', 'STRLIT'):
                    return node.value
                else:
                    raise DebugSequenceSemanticError(f"unexpected literal type {node.type}")
            elif isinstance(node, int):
                return node
            else:
                raise DebugSequenceSemanticError("unexpected node type when expecting atom")

        def _format_atom(self, node: NodeType) -> str:
            """@brief Format an atom for trace logging."""
            if isinstance(node, LarkToken):
                if node.type == 'IDENT':
                    try:
                        return node.value + "{" + hex(self._scope.get(node.value)) + "}"
                    except KeyError:
                        TRACE.debug("reference to undefined variable %s... %s", node.value, self._scope.dump())
                        return node.value + "{undefined}"
                elif node.type == 'INTLIT':
                    return hex(node.value)
                elif node.type == 'STRLIT':
                    return f"'{node.value}'"
                else:
                    raise DebugSequenceSemanticError(f"unexpected literal type {node.type}")
            elif isinstance(node, int):
                return hex(node)
            else:
                return f"?<{type(node).__class__}>?"

    def __init__(self, tree: LarkTree, scope: Scope, context: DebugSequenceExecutionContext) -> None:
        """@brief Constructor.

        The provided AST is semantically checked and optimized.

        @param self This interpreter.
        @param tree The abstract syntax tree to interpret.
        @param scope Scope within which the AST will execute.
        @param delegate Delegate providing debug sequence function implementations.

        @exception DebugSequenceSemanticError A semantic error was discovered in the provided code.
        """
        super().__init__()
        self._scope = scope
        self._context = context

        # First run the semantic checker, so semantic errors are raised prior to actually
        # performing any actions.
        checker = SemanticChecker(tree, self._scope, context)
        checker.check()

        # Do some optimization.
        self._tree = _ConstantFolder().transform(tree)

    def execute(self) -> int:
        """@brief Runs the statements in the AST passed to the constructor.
        @return The value of the last statement is returned to the caller.
        """
        visitor = self._InterpreterVisitor(self._scope, self._context)
        return visitor.visit(self._tree)




# pyOCD debugger
# Copyright (c) 2020 Arm Limited
# Copyright (c) 2021 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from time import sleep
from ...flash.flash import Flash
from ...coresight.coresight_target import CoreSightTarget
from ...core.memory_map import (FlashRegion, RamRegion, MemoryMap)
from ...core.target import Target
from ...coresight.cortex_m import CortexM
from ...core import exceptions
from ...utility.timeout import Timeout

LOG = logging.getLogger(__name__)

flash_algo = {
    'load_address': 0x00100000,

    # Flash algorithm as a hex string
    'instructions': [
        0xE00ABE00, 0x062D780D, 0x24084068, 0xD3000040, 0x1E644058, 0x1C49D1FA, 0x2A001E52, 0x4770D1F2,
        0x280a49fa, 0x698ad10a, 0x7f80f012, 0x698ad1fb, 0x0f7ff412, 0x220dd1fb, 0x2020f881, 0xf012698a,
        0xd1fb7f80, 0xf412698a, 0xd1fb0f7f, 0x0020f881, 0xb1284770, 0x70116802, 0x1c496801, 0x47706001,
        0x48eab2ca, 0xd10a2a0a, 0xf0116981, 0xd1fb7f80, 0xf4116981, 0xd1fb0f7f, 0xf880210d, 0x69811020,
        0x7f80f011, 0x6981d1fb, 0x0f7ff411, 0xf880d1fb, 0x47702020, 0x41f0e92d, 0x460e1e14, 0xf04f4680,
        0xf04f0500, 0xdd100720, 0x21007832, 0xb1224630, 0x2f01f810, 0x2a001c49, 0x42a1d1fa, 0x2400bfac,
        0xf0131a64, 0xbf180f02, 0xf0132730, 0xd1090f01, 0xdd072c00, 0x46404639, 0xffbbf7ff, 0x1c6d1e64,
        0xdcf72c00, 0xb1407830, 0x4640b2c1, 0xffb1f7ff, 0x0f01f816, 0x28001c6d, 0x2c00d1f6, 0x4639dd07,
        0xf7ff4640, 0x1e64ffa6, 0x2c001c6d, 0x4628dcf7, 0x81f0e8bd, 0x45f0e92d, 0x469cb083, 0xe9dd4680,
        0x2000730b, 0x46059e0a, 0xb1294682, 0x0f00f1bc, 0x2a0ad016, 0xe013d00e, 0xf88d2030, 0xf88d0000,
        0x463ba001, 0x46694632, 0xf7ff4640, 0xb003ffa3, 0x85f0e8bd, 0x0c00f1b1, 0x2001bfbc, 0x0100f1cc,
        0x040bf10d, 0xa00bf88d, 0xfbb1b189, 0xfb02fcf2, 0xf1bc1c1c, 0xbfa40f0a, 0xf1ac449c, 0xf10c0c3a,
        0xfbb10c30, 0xf804f1f2, 0x2900cd01, 0xb178d1ed, 0xbf182e00, 0x0f02f017, 0xf04fd007, 0x4640012d,
        0xff57f7ff, 0x1e761c6d, 0x202de002, 0x0d01f804, 0x4632463b, 0x46404621, 0xff6cf7ff, 0x4428b003,
        0x85f0e8bd, 0xe92db40f, 0xb0844df0, 0x9c0c2700, 0xad0d463e, 0x28007820, 0xf04fd075, 0xf04f0b41,
        0x46ba0861, 0x2825b2c0, 0x2200d17b, 0x0f01f814, 0x28004613, 0x2825d07f, 0x282dd073, 0x2301bf04,
        0x78201c64, 0xd1052830, 0x0f01f814, 0x0302f043, 0xd0f92830, 0x3830b2c0, 0xd80a2809, 0x0082eb02,
        0xf8140040, 0x38301b01, 0x7820180a, 0x28093830, 0x7820d9f4, 0xd00a2873, 0xd0112864, 0xd01c2878,
        0xd0272858, 0xd0322875, 0xd03f2863, 0xf855e04e, 0x29001b04, 0xa16ebf08, 0xf7ff4638, 0xe024ff1b,
        0x2300e9cd, 0x8008f8cd, 0x1b04f855, 0x220a2301, 0xf7ff4638, 0x4406ff4f, 0xe9cde038, 0xf8cd2300,
        0xf8558008, 0x23001b04, 0x46382210, 0xff42f7ff, 0xe02b4406, 0x2300e9cd, 0xb008f8cd, 0x1b04f855,
        0x22102300, 0xf7ff4638, 0x4406ff35, 0xe9cde01e, 0xf8cd2300, 0xf8558008, 0x23001b04, 0x4638220a,
        0xff28f7ff, 0xe01be7f1, 0xe014e00b, 0x0b04f815, 0x000cf88d, 0xa00df88d, 0x4638a903, 0xfedaf7ff,
        0xb2c1e7e3, 0xf7ff4638, 0x1c76feb4, 0x0f01f814, 0xf47f2800, 0x2f00af77, 0x6838bf1c, 0xa000f880,
        0xb0044630, 0x0df0e8bd, 0xfb14f85d, 0xea236803, 0x43110101, 0x47706001, 0x42812100, 0x1c49bfb8,
        0x4770dbfb, 0xf44f493c, 0xf84140c6, 0x60480f70, 0x60c86088, 0x61486108, 0x21014838, 0x21006741,
        0x49376781, 0x21646041, 0x1c402000, 0xdbfc4288, 0xf04f4770, 0x48334202, 0xf8c24934, 0x48320100,
        0x49336008, 0x60081200, 0x12001d09, 0x49316008, 0x1d096008, 0x3001f04f, 0x210a6008, 0x1c402000,
        0xdbfc4288, 0x2136482c, 0x21106241, 0x21706281, 0xf24062c1, 0x63013101, 0x60012155, 0x210a4827,
        0x0100f8c2, 0x1c402000, 0xdbfc4288, 0xf3bf4770, 0xf3bf8f6f, 0x49228f4f, 0x60082000, 0x8f6ff3bf,
        0x8f4ff3bf, 0x481f4920, 0x47706008, 0x8f6ff3bf, 0x8f4ff3bf, 0x21e0f04f, 0x61082000, 0x8f6ff3bf,
        0x8f4ff3bf, 0xf3bf4770, 0xf3bf8f6f, 0x20008f4f, 0xf1004601, 0x1d0022e0, 0x1100f8c2, 0xdbf82820,
        0x8f6ff3bf, 0x8f4ff3bf, 0x00004770, 0x83015000, 0x6c756e28, 0x0000296c, 0x82021000, 0x85020000,
        0x8660061a, 0xc1900d01, 0x01000001, 0x82000a04, 0x83000a00, 0x85000a00, 0x85024000, 0xc1900d11,
        0xe000ef50, 0x00040200, 0xe000ed14, 0x20e0f04f, 0xf8402100, 0x4ab31f10, 0x7100f04f, 0x17516011,
        0x1170f8c0, 0x1174f8c0, 0x1270f8c0, 0x1274f8c0, 0x4bad4770, 0xf0406818, 0x60180002, 0x680849ab,
        0x0001f040, 0x49aa6008, 0xf0406808, 0x60080001, 0xf8d149a8, 0xb1780118, 0xf8c12200, 0x200f2100,
        0x0104f8c1, 0x0108f8d1, 0xd1fb280f, 0x2104f8c1, 0x0108f8d1, 0xd1fb2800, 0x20002264, 0x42901c40,
        0xf8d1dbfc, 0xf0200110, 0xf8c10003, 0xf8d10110, 0xf0200118, 0xf8c10003, 0xf8d10118, 0xf0200114,
        0xf8c10003, 0x68180114, 0x0002f020, 0x47706018, 0x21004891, 0x68016201, 0x0104f021, 0x47706001,
        0x45f0e92d, 0xf04f488c, 0xf8c00c00, 0x6801c020, 0x0104f021, 0xf7ff6001, 0xf3bfff1c, 0xf3bf8f6f,
        0x48868f4f, 0xc000f8c0, 0x8f6ff3bf, 0x8f4ff3bf, 0x48834984, 0xf3bf6008, 0xf3bf8f6f, 0xf04f8f4f,
        0xf3bf20e0, 0xf3bf8f6f, 0xf8c08f4f, 0x4a75c010, 0x7100f04f, 0x17516011, 0x1180f8c0, 0x1184f8c0,
        0x1280f8c0, 0x1284f8c0, 0x8f6ff3bf, 0x8f4ff3bf, 0xf1002000, 0x1d0021e0, 0xc100f8c1, 0xdbf82820,
        0x8f6ff3bf, 0x8f4ff3bf, 0xff73f7ff, 0x81bcf8df, 0x486d2201, 0xf8c84b6e, 0xf2472038, 0xf8c03101,
        0xf8c3c0c4, 0xf8c31130, 0xe9c01134, 0x26032c02, 0xe9c02406, 0xe9c04c04, 0xf243c206, 0xf04f3785,
        0xe9c04502, 0xf8d56700, 0xf0155100, 0xf04f4f00, 0xbf190504, 0x6aa1f44f, 0xa20ae9c0, 0x0aa8f04f,
        0xa50ae9c0, 0x0a07f04f, 0xc038f8c0, 0x20c4f8c0, 0x2038f8c8, 0xc0c4f8c0, 0x1130f8c3, 0x1134f8c3,
        0x60476006, 0x2121f240, 0xf8c06081, 0x6104c00c, 0xc014f8c0, 0xc018f8c0, 0x216a61c2, 0x62c46281,
        0x63456305, 0xc03cf8c0, 0xa008f8c0, 0x20002164, 0x42881c40, 0xf44fdbfc, 0x671840c6, 0x67986758,
        0xf8c367d8, 0xf8c30080, 0x48420084, 0x60414942, 0x2c1de9c0, 0x20002164, 0x42881c40, 0x2000dbfc,
        0x85f0e8bd, 0xb510493b, 0x6048483c, 0xf7ffa03c, 0x2000fda9, 0xb510bd10, 0xf7ffa03b, 0x4b3dfda3,
        0xf44f4934, 0x20001480, 0x0cfff04f, 0x20dcf891, 0x0f01f012, 0xeb03d1fa, 0xf1a23200, 0x610a4280,
        0xc05ef881, 0x20dcf891, 0x0f01f012, 0x1c40d1fa, 0x3f14ebb0, 0xa030d3ea, 0xfd84f7ff, 0xbd102000,
        0x4c24b570, 0x68624605, 0xa02d4601, 0xfd7af7ff, 0x20dcf894, 0x0f01f012, 0xf1a5d1fa, 0x61204080,
        0xf88420ff, 0xf894005e, 0xf01000dc, 0xd1fa0001, 0xb570bd70, 0x44a0f100, 0x46154816, 0x6843460e,
        0x4621460a, 0xf7ffa028, 0x1cf0fd5d, 0x0003f030, 0xf855d005, 0xf8441b04, 0x1f001b04, 0x2000d1f9,
        0x0000bd70, 0xe000ed04, 0x82020460, 0x82020500, 0x82020520, 0x82020000, 0x83011000, 0xe000ef50,
        0x00040200, 0xe000ed14, 0x83015000, 0x85041000, 0x82021000, 0x85020000, 0x8660061a, 0x0660860a,
        0x6e696e55, 0x000a7469, 0x53415245, 0x48432045, 0x000a5049, 0x406f4000, 0x454e4f44, 0x0000000a,
        0x20337773, 0x53415245, 0x45532045, 0x524f5443, 0x7830202c, 0x202c7825, 0x73616c66, 0x78305b68,
        0x0a5d7825, 0x00000000, 0x676f7250, 0x206d6172, 0x3d726461, 0x78257830, 0x7a73202c, 0x2578303d,
        0x66202c78, 0x6873616c, 0x2578305b, 0x000a5d78, 0x00000000
    ],

    # Relative function addresses
    'pc_init': 0x00100501,
    'pc_unInit': 0x00100665,
    'pc_program_page': 0x001006f3,
    'pc_erase_sector': 0x001006c1,
    'pc_eraseAll': 0x00100677,

    'static_base': 0x001007d0,
    'begin_stack': 0x00100a00,
    'begin_data': 0x00100000 + 0x1000,
    'page_size': 0x400,
    'analyzer_supported': False,
    'analyzer_address': 0x00000000,
    # 'page_buffers': [0x00101000, 0x00101400],  # Enable double buffering
    'min_program_length': 0x400

}


class Flash_s5js100(Flash):
    def __init__(self, target, flash_algo):
        super(Flash_s5js100, self).__init__(target, flash_algo)
        self._did_prepare_target = False
        # LOG.info("S5JS100.Flash_s5js100.__init__ c")

    def init(self, operation, address=None, clock=0, reset=True):
        # LOG.info("S5JS100.Flash_s5js100.init c")

        if self._active_operation != operation and self._active_operation is not None:
            self.uninit()

        super(Flash_s5js100, self).init(operation, address, clock, reset)

    def uninit(self):
        # LOG.info("S5JS100.Flash_s5js100.uninit c")
        if self._active_operation is None:
            return

        super(Flash_s5js100, self).uninit()


ERASE_ALL_WEIGHT = 140  # Time it takes to perform a chip erase
ERASE_SECTOR_WEIGHT = 1  # Time it takes to erase a page
# Time it takes to program a page (Not including data transfer time)
PROGRAM_PAGE_WEIGHT = 1


class S5JS100(CoreSightTarget):
    VENDOR = "Samsung"
    AP_NUM = 0
    ROM_ADDR = 0xE00FE000

    MEMORY_MAP = MemoryMap(
        FlashRegion(start=0x406f4000, length=0x00100000,
                    page_size=0x400, blocksize=0x1000,
                    is_boot_memory=True,
                    erased_byte_value=0xFF,
                    algo=flash_algo,
                    erase_all_weight=ERASE_ALL_WEIGHT,
                    erase_sector_weight=ERASE_SECTOR_WEIGHT,
                    program_page_weight=PROGRAM_PAGE_WEIGHT,
                    flash_class=Flash_s5js100),
        RamRegion(start=0x00100000, length=0x80000)
    )

    def __init__(self, session):
        super(S5JS100, self).__init__(session, self.MEMORY_MAP)
        self.AP_NUM = 0

    def create_init_sequence(self):
        seq = super(S5JS100, self).create_init_sequence()
        seq.wrap_task(
            'discovery', lambda seq: seq.replace_task(
                'find_aps', self.find_aps).replace_task(
                'create_cores', self.create_s5js100_core).insert_before(
                'find_components', ('fixup_ap_base_addrs', self._fixup_ap_base_addrs), ))
        return seq

    def _fixup_ap_base_addrs(self):
        self.dp.aps[self.AP_NUM].rom_addr = self.ROM_ADDR

    def find_aps(self):
        if self.dp.valid_aps is not None:
            return

        self.dp.valid_aps = (self.AP_NUM,)

    def create_s5js100_core(self):
        core = CortexM_S5JS100(
            self.session, self.aps[self.AP_NUM], self.memory_map, 0)
        core.default_reset_type = self.ResetType.SW
        self.aps[self.AP_NUM].core = core
        core.init()
        self.add_core(core)


class CortexM_S5JS100(CortexM):

    def reset(self, reset_type=None):
        # Always use software reset for S5JS100 since the hardware version
        self.session.notify(Target.Event.PRE_RESET, self)

        # LOG.info("s5js100 reset HW")
        self.S5JS100_reset_type = reset_type
        self.write_memory(0x82020018, 0x1 << 1)
        self.write_memory(0x83011000, 0x4 << 0)  # enable watchdog
        self.write_memory(0x8301100c, 0x1 << 0)
        self.write_memory(0x83011010, 0x1 << 0)
        self.write_memory(0x83011020, 0x1 << 0)
        self.write_memory(0x83011800, 0x1 << 0)  # clock gating disable
        # set 1s to be reset , 1 sec=32768
        self.write_memory(0x83011004, 32768 << 0)
        # force to load value to be reset
        self.write_memory(0x83011008, 0xFF << 0)

        xpsr = self.read_core_register('xpsr')
        if xpsr & self.XPSR_THUMB == 0:
            self.write_core_register('xpsr', xpsr | self.XPSR_THUMB)

        self.write_memory(0x120000, 0xe7fe)
        self.write_core_register('pc', 0x120000)

        self.flush()
        self.resume()
        with Timeout(5.0) as t_o:
            while t_o.check():
                try:
                    dhcsr_reg = self.read32(CortexM.DHCSR)
                    LOG.debug("reg = %x", dhcsr_reg)
                    if (dhcsr_reg & CortexM.S_RESET_ST) != 0:
                        break
                    sleep(0.1)
                except exceptions.TransferError:
                    self.flush()
                    self._ap.dp.connect()
                    sleep(0.01)
            else:
                raise exceptions.TimeoutError("Timeout waiting for reset")
        self.session.notify(Target.Event.POST_RESET, self)

    def reset_and_halt(self, reset_type=None):
        # LOG.info("reset_and_halt")
        reset_catch_saved_demcr = self.read_memory(CortexM.DEMCR)
        if (reset_catch_saved_demcr & CortexM.DEMCR_VC_CORERESET) == 0:
            self.write_memory(
                CortexM.DEMCR,
                reset_catch_saved_demcr | CortexM.DEMCR_VC_CORERESET)
        self.reset(reset_type)
        sleep(0.1)
        self.halt()
        self.wait_halted()
        self.write_memory(CortexM.DEMCR, reset_catch_saved_demcr)

    def wait_halted(self):
        with Timeout(5.0) as t_o:
            while t_o.check():
                try:
                    if not self.is_running():
                        break
                except exceptions.TransferError:
                    self.flush()
                    sleep(0.01)
            else:
                raise exceptions.TimeoutError("Timeout waiting for target halt")

    def get_state(self):
        # LOG.info("s5js100.get_state")
        try:
            dhcsr = self.read_memory(CortexM.DHCSR)
            # LOG.info("s5js100.get_state dhcsr 0x%x", dhcsr)
        except exceptions.TransferError:
            # LOG.info("s5js100.get_state read fail dhcsr..try more")
            self._ap.dp.connect()
            dhcsr = self.read_memory(CortexM.DHCSR)
            # LOG.info("fail s5js100.get_state dhcsr 0x%x", dhcsr)

        if dhcsr & CortexM.S_RESET_ST:
            # Reset is a special case because the bit is sticky and really means
            # "core was reset since last read of DHCSR". We have to re-read the
            # DHCSR, check if S_RESET_ST is still set and make sure no instructions
            # were executed by checking S_RETIRE_ST.
            newDhcsr = self.read_memory(CortexM.DHCSR)
            if (newDhcsr & CortexM.S_RESET_ST) and not (
                    newDhcsr & CortexM.S_RETIRE_ST):
                return Target.State.RESET
        if dhcsr & CortexM.S_LOCKUP:
            return Target.State.LOCKUP
        elif dhcsr & CortexM.S_SLEEP:
            return Target.State.SLEEPING
        elif dhcsr & CortexM.S_HALT:
            return Target.State.HALTED
        else:
            return Target.State.RUNNING

    def set_breakpoint(self, addr, type=Target.BreakpointType.HW):
        # s5js100 don't support Target.BreakpointType.SW
        return super(
            CortexM_S5JS100,
            self).set_breakpoint(
            addr,
            Target.BreakpointType.HW)



# pyOCD debugger
# Copyright (c) 2015-2021 Arm Limited
# Copyright (c) 2021 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import (Callable, Dict, Optional, NamedTuple, Tuple, Union, TYPE_CHECKING)

from .ap import (APAddressBase, AccessPort)
from .cortex_m import CortexM
from .cortex_m_v8m import CortexM_v8M
from .fpb import FPB
from .dwt import (DWT, DWTv2)
from .itm import ITM
from .tpiu import TPIU
from .gpr import GPR
from .sdc600 import SDC600
from .funnel import TraceFunnel

if TYPE_CHECKING:
    from .component import CoreSightComponent
    from .rom_table import CoreSightComponentID
    from ..core.memory_interface import MemoryInterface
    from .dap import DebugPort

# Component classes.
ROM_TABLE_CLASS = 0x1
CORESIGHT_CLASS = 0x9
GENERIC_CLASS = 0xe
SYSTEM_CLASS = 0xf # CoreLink, PrimeCell, or other system component with no standard register layout.

#  [11:8] continuation
#  [6:0]  ID
ARM_ID = 0x43b
ARM_CHINA_ID = 0xa75
FSL_ID = 0x00e
STM_ID = 0x020

## Map of JEP106 IDs to vendor name.
VENDOR_NAMES_MAP: Dict[int, str] = {
    ARM_ID: "Arm",
    ARM_CHINA_ID: "Arm China",
    FSL_ID: "NXP",
    STM_ID: "ST",
}

# CoreSight devtype
#  Major Type [3:0]
#  Minor Type [7:4]
#
# CoreSight Major Types
#  0 = Miscellaneous
#  1 = Trace Sink
#  2 = Trace Link
#  3 = Trace Source
#  4 = Debug Control
#  5 = Debug Logic
#
# Known devtype values
#  0x11 = TPIU
#  0x21 = ETB
#  0x31 = MTB
#  0x12 = Trace funnel (CSFT)
#  0x32 = TMC
#  0x13 = CPU trace source (ETM, MTB?)
#  0x43 = ITM
#  0x14 = ECT/CTI/CTM
#  0x34 = Granular Power Requestor
#  0x15 = CPU debug
#  0x16 = PMU

# Two factory signatures are supported. The discovery classes ensure the right one is called by
# filtering APs as appropriate.
ComponentFactory = Callable[["MemoryInterface", "CoreSightComponentID", int], "CoreSightComponent"]
APFactory = Callable[["DebugPort", APAddressBase, Optional["CoreSightComponentID"]], AccessPort]

class CmpInfo(NamedTuple):
    """@brief Combines a component and product name with a factory method."""
    name: str
    product: Optional[str]
    factory: Optional[Union[ComponentFactory, APFactory]]

## Map from (designer, class, part, devtype, archid) to component name, product name, and factory.
COMPONENT_MAP: Dict[Tuple[int, int, Optional[int], Optional[int], int], CmpInfo] = {
  # Archid-only entries
  # Designer|Component Class |Part  |Type |Archid           |Name              |Product    |Factory
    (ARM_ID, CORESIGHT_CLASS, None,  None, 0x0a00) : CmpInfo('RASv1',           None,       None                ),
    (ARM_ID, CORESIGHT_CLASS, None,  None, 0x1a01) : CmpInfo('ITMv2',           None,       ITM.factory         ),
    (ARM_ID, CORESIGHT_CLASS, None,  None, 0x1a02) : CmpInfo('DWTv2',           None,       DWTv2.factory       ),
    (ARM_ID, CORESIGHT_CLASS, None,  None, 0x1a03) : CmpInfo('FPBv2',           None,       FPB.factory         ),
    (ARM_ID, CORESIGHT_CLASS, None,  None, 0x2a04) : CmpInfo('v8-M Debug',      None,       CortexM_v8M.factory ),
    (ARM_ID, CORESIGHT_CLASS, None,  None, 0x6a05) : CmpInfo('v8-R Debug',      None,       None                ),
    (ARM_ID, CORESIGHT_CLASS, None,  None, 0x0a06) : CmpInfo('v8-M PMUv1',      None,       None                ),
    (ARM_ID, CORESIGHT_CLASS, None,  None, 0x4a13) : CmpInfo('ETMv4',           None,       None                ),
    (ARM_ID, CORESIGHT_CLASS, None,  None, 0x1a14) : CmpInfo('CTIv2',           None,       None                ),
    (ARM_ID, CORESIGHT_CLASS, None,  None, 0x6a15) : CmpInfo('v8.0-A Debug',    None,       None                ),
    (ARM_ID, CORESIGHT_CLASS, None,  None, 0x7a15) : CmpInfo('v8.1-A Debug',    None,       None                ),
    (ARM_ID, CORESIGHT_CLASS, None,  None, 0x8a15) : CmpInfo('v8.2-A Debug',    None,       None                ),
    (ARM_ID, CORESIGHT_CLASS, None,  None, 0x2a16) : CmpInfo('PMUv2',           None,       None                ),
    (ARM_ID, CORESIGHT_CLASS, None,  None, 0x0a17) : CmpInfo('MEM-APv2',        None,       AccessPort.create   ),
    (ARM_ID, CORESIGHT_CLASS, None,  None, 0x0a34) : CmpInfo('GPR',             None,       GPR.factory         ),
    (ARM_ID, CORESIGHT_CLASS, None,  None, 0x0a55) : CmpInfo('PMCv0/1',         None,       None                ),
    (ARM_ID, CORESIGHT_CLASS, None,  None, 0x2a56) : CmpInfo('SMMUv3',          None,       None                ),
    (ARM_ID, CORESIGHT_CLASS, None,  None, 0x0a63) : CmpInfo('STMv1',           None,       None                ),
    (ARM_ID, CORESIGHT_CLASS, None,  None, 0x0a66) : CmpInfo('AMUv1',           None,       None                ),
    (ARM_ID, CORESIGHT_CLASS, None,  None, 0x0a75) : CmpInfo('ELA',             None,       None                ),
    (ARM_ID, CORESIGHT_CLASS, None,  None, 0x0af7) : CmpInfo('ROM',             None,       None                ),
  # Full ID entries
  # Designer|Component Class |Part  |Type |Archid           |Name              |Product    |Factory
    (ARM_ID, CORESIGHT_CLASS, 0x193, 0x00, 0x0a57) : CmpInfo('TSGEN',           'CS-600',   None                ),
    (ARM_ID, CORESIGHT_CLASS, 0x906, 0x14, 0)      : CmpInfo('CTI',             'CS-400',   None                ),
    (ARM_ID, CORESIGHT_CLASS, 0x907, 0x21, 0)      : CmpInfo('ETB',             'CS-400',   None                ),
    (ARM_ID, CORESIGHT_CLASS, 0x908, 0x12, 0)      : CmpInfo('Trace Funnel',    'CS-400',   TraceFunnel.factory ),
    (ARM_ID, CORESIGHT_CLASS, 0x909, 0x22, 0)      : CmpInfo('Trace Replicator',None,       None                ),
    (ARM_ID, CORESIGHT_CLASS, 0x912, 0x11, 0)      : CmpInfo('TPIU',            'CS-400',   TPIU.factory        ),
    (ARM_ID, CORESIGHT_CLASS, 0x913, 0x43, 0)      : CmpInfo('ITM',             'CS-400',   None                ),
    (ARM_ID, CORESIGHT_CLASS, 0x914, 0x11, 0)      : CmpInfo('SWO',             'CS-400',   TPIU.factory        ),
    (ARM_ID, CORESIGHT_CLASS, 0x917, 0x43, 0)      : CmpInfo('HTM',             'CS-400',   None                ),
    (ARM_ID, CORESIGHT_CLASS, 0x923, 0x11, 0)      : CmpInfo('TPIU',            'M3',       TPIU.factory        ),
    (ARM_ID, CORESIGHT_CLASS, 0x924, 0x13, 0)      : CmpInfo('ETM',             'M3',       None                ),
    (ARM_ID, CORESIGHT_CLASS, 0x925, 0x13, 0)      : CmpInfo('ETM',             'M4',       None                ),
    (ARM_ID, CORESIGHT_CLASS, 0x926, 0x13, 0)      : CmpInfo('ETM',             'SC300',    None                ),
    (ARM_ID, CORESIGHT_CLASS, 0x927, 0x11, 0)      : CmpInfo('TPIU',            'SC300',    None                ),
    (ARM_ID, CORESIGHT_CLASS, 0x932, 0x31, 0x0a31) : CmpInfo('MTB',             'M0+',      None                ),
    (ARM_ID, CORESIGHT_CLASS, 0x950, 0x13, 0)      : CmpInfo('PTM',             'A9',       None                ),
    (ARM_ID, CORESIGHT_CLASS, 0x961, 0x32, 0)      : CmpInfo('ETF',             None,       None                ), # Trace Memory Controller ETF
    (ARM_ID, CORESIGHT_CLASS, 0x962, 0x63, 0x0a63) : CmpInfo('STM',             None,       None                ), # System Trace Macrocell
    (ARM_ID, CORESIGHT_CLASS, 0x962, 0x63, 0)      : CmpInfo('STM',             None,       None                ), # System Trace Macrocell (archid=0)
    (ARM_ID, CORESIGHT_CLASS, 0x963, 0x63, 0x0a63) : CmpInfo('STM-500',         None,       None                ), # System Trace Macrocell
    (ARM_ID, CORESIGHT_CLASS, 0x975, 0x13, 0x4a13) : CmpInfo('ETM',             'M7',       None                ),
    (ARM_ID, CORESIGHT_CLASS, 0x9a0, 0x16, 0)      : CmpInfo('PMU',             'A9',       None                ),
    (ARM_ID, CORESIGHT_CLASS, 0x9a1, 0x11, 0)      : CmpInfo('TPIU',            'M4',       TPIU.factory        ),
    (ARM_ID, CORESIGHT_CLASS, 0x9a3, 0x13, 0x0)    : CmpInfo('MTB',             'M0',       None                ),
    (ARM_ID, CORESIGHT_CLASS, 0x9a4, 0x34, 0x0a34) : CmpInfo('GPR',             None,       GPR.factory         ), # Granular Power Requestor
    (ARM_ID, CORESIGHT_CLASS, 0x9a5, 0x16, 0)      : CmpInfo('PMU',             'A5',       None                ),
    (ARM_ID, CORESIGHT_CLASS, 0x9a6, 0x14, 0x1a14) : CmpInfo('CTI',             'M0+',      None                ),
    (ARM_ID, CORESIGHT_CLASS, 0x9a7, 0x16, 0)      : CmpInfo('PMU',             'A7',       None                ),
    (ARM_ID, CORESIGHT_CLASS, 0x9a9, 0x11, 0)      : CmpInfo('TPIU',            'M7',       TPIU.factory        ),
    (ARM_ID, CORESIGHT_CLASS, 0x9ba, 0x55, 0x0a55) : CmpInfo('PMC-100',         None,       None                ), # Programmable MBIST Controller
    (ARM_ID, CORESIGHT_CLASS, 0x9db, 0x13, 0x4a13) : CmpInfo('ETM',             'A32',      None                ), # ETMv4
    (ARM_ID, CORESIGHT_CLASS, 0x9db, 0x14, 0x1a14) : CmpInfo('CTI',             'A32',      None                ), # CTIv2
    (ARM_ID, CORESIGHT_CLASS, 0x9db, 0x16, 0x2a16) : CmpInfo('PMU',             'A32',      None                ), # PMUv3
    (ARM_ID, CORESIGHT_CLASS, 0x9e2, 0x00, 0x0a17) : CmpInfo('APB-AP',          'CS-600',   AccessPort.create   ),
    (ARM_ID, CORESIGHT_CLASS, 0x9e3, 0x00, 0x0a17) : CmpInfo('AHB-AP',          'CS-600',   AccessPort.create   ),
    (ARM_ID, CORESIGHT_CLASS, 0x9e4, 0x00, 0x0a17) : CmpInfo('AXI-AP',          'CS-600',   AccessPort.create   ),
    (ARM_ID, CORESIGHT_CLASS, 0x9e5, 0x00, 0x0a47) : CmpInfo('APv1 Adapter',    'CS-600',   AccessPort.create   ),
    (ARM_ID, CORESIGHT_CLASS, 0x9e6, 0x00, 0x0a27) : CmpInfo('JTAG-AP',         'CS-600',   AccessPort.create   ),
    (ARM_ID, CORESIGHT_CLASS, 0x9e7, 0x11, 0)      : CmpInfo('TPIU',            'CS-600',   TPIU.factory        ),
    (ARM_ID, CORESIGHT_CLASS, 0x9e8, 0x21, 0)      : CmpInfo('ETR',             'CS-600',   None                ),
    (ARM_ID, CORESIGHT_CLASS, 0x9e9, 0x21, 0)      : CmpInfo('ETB',             'CS-600',   None                ),
    (ARM_ID, CORESIGHT_CLASS, 0x9ea, 0x32, 0)      : CmpInfo('ETF',             'CS-600',   None                ),
    (ARM_ID, CORESIGHT_CLASS, 0x9eb, 0x12, 0)      : CmpInfo('ATB Funnel',      'CS-600',   TraceFunnel.factory ),
    (ARM_ID, CORESIGHT_CLASS, 0x9ec, 0x22, 0)      : CmpInfo('ATB Replicator',  'CS-600',   None                ),
    (ARM_ID, CORESIGHT_CLASS, 0x9ed, 0x14, 0x1a14) : CmpInfo('CTI',             'CS-600',   None                ),
    (ARM_ID, CORESIGHT_CLASS, 0x9ee, 0x00, 0)      : CmpInfo('CATU',            'CS-600',   None                ),
    (ARM_ID, CORESIGHT_CLASS, 0x9ef, 0x00, 0x0a57) : CmpInfo('SDC-600',         None,       SDC600.factory      ),
    (ARM_ID, CORESIGHT_CLASS, 0x9f0, 0x00, 0)      : CmpInfo('GPIO Control',    None,       None                ),
    (ARM_ID, CORESIGHT_CLASS, 0xc05, 0x15, 0)      : CmpInfo('CPU',             'A5',       None                ),
    (ARM_ID, CORESIGHT_CLASS, 0xc07, 0x15, 0)      : CmpInfo('CPU',             'A7',       None                ),
    (ARM_ID, CORESIGHT_CLASS, 0xc08, 0x15, 0)      : CmpInfo('CPU',             'A8',       None                ),
    (ARM_ID, CORESIGHT_CLASS, 0xc09, 0x15, 0)      : CmpInfo('CPU',             'A9',       None                ),
    (ARM_ID, CORESIGHT_CLASS, 0xc0d, 0x15, 0)      : CmpInfo('CPU',             'A12',      None                ),
    (ARM_ID, CORESIGHT_CLASS, 0xc0e, 0x15, 0)      : CmpInfo('CPU',             'A17',      None                ),
    (ARM_ID, CORESIGHT_CLASS, 0xc0f, 0x15, 0)      : CmpInfo('CPU',             'A15',      None                ),
    (ARM_ID, CORESIGHT_CLASS, 0xd01, 0x15, 0x6a15) : CmpInfo('CPU',             'A32',      None                ),
    (ARM_ID, CORESIGHT_CLASS, 0xd02, 0x15, 0x6a15) : CmpInfo('CPU',             'A34',      None                ),
    (ARM_ID, CORESIGHT_CLASS, 0xd03, 0x15, 0x6a15) : CmpInfo('CPU',             'A53',      None                ),
    (ARM_ID, CORESIGHT_CLASS, 0xd04, 0x15, 0x6a15) : CmpInfo('CPU',             'A35',      None                ),
    (ARM_ID, CORESIGHT_CLASS, 0xd05, 0x15, 0x6a15) : CmpInfo('CPU',             'A55',      None                ),
    (ARM_ID, CORESIGHT_CLASS, 0xd06, 0x15, 0x6a15) : CmpInfo('CPU',             'A65',      None                ),
    (ARM_ID, CORESIGHT_CLASS, 0xd07, 0x15, 0x6a15) : CmpInfo('CPU',             'A57',      None                ),
    (ARM_ID, CORESIGHT_CLASS, 0xd08, 0x15, 0x6a15) : CmpInfo('CPU',             'A72',      None                ),
    (ARM_ID, CORESIGHT_CLASS, 0xd09, 0x15, 0x6a15) : CmpInfo('CPU',             'A73',      None                ),
    (ARM_ID, CORESIGHT_CLASS, 0xd0a, 0x15, 0x6a15) : CmpInfo('CPU',             'A75',      None                ),
    (ARM_ID, CORESIGHT_CLASS, 0xd0b, 0x15, 0x6a15) : CmpInfo('CPU',             'A76',      None                ),
    (ARM_ID, CORESIGHT_CLASS, 0xd20, 0x11, 0)      : CmpInfo('TPIU',            'M23',      TPIU.factory        ),
    (ARM_ID, CORESIGHT_CLASS, 0xd20, 0x13, 0)      : CmpInfo('ETM',             'M23',      None                ),
    (ARM_ID, CORESIGHT_CLASS, 0xd20, 0x31, 0x0a31) : CmpInfo('MTB',             'M23',      None                ),
    (ARM_ID, CORESIGHT_CLASS, 0xd20, 0x00, 0x1a02) : CmpInfo('DWT',             'M23',      DWTv2.factory       ),
    (ARM_ID, CORESIGHT_CLASS, 0xd20, 0x00, 0x1a03) : CmpInfo('BPU',             'M23',      FPB.factory         ),
    (ARM_ID, CORESIGHT_CLASS, 0xd20, 0x14, 0x1a14) : CmpInfo('CTI',             'M23',      None                ),
    (ARM_ID, CORESIGHT_CLASS, 0xd20, 0x00, 0x2a04) : CmpInfo('SCS',             'M23',      CortexM_v8M.factory ),
    (ARM_ID, CORESIGHT_CLASS, 0xd21, 0x31, 0x0a31) : CmpInfo('MTB',             'M33',      None                ),
    (ARM_ID, CORESIGHT_CLASS, 0xd21, 0x43, 0x1a01) : CmpInfo('ITM',             'M33',      ITM.factory         ),
    (ARM_ID, CORESIGHT_CLASS, 0xd21, 0x00, 0x1a02) : CmpInfo('DWT',             'M33',      DWTv2.factory       ),
    (ARM_ID, CORESIGHT_CLASS, 0xd21, 0x00, 0x1a03) : CmpInfo('BPU',             'M33',      FPB.factory         ),
    (ARM_ID, CORESIGHT_CLASS, 0xd21, 0x14, 0x1a14) : CmpInfo('CTI',             'M33',      None                ),
    (ARM_ID, CORESIGHT_CLASS, 0xd21, 0x00, 0x2a04) : CmpInfo('SCS',             'M33',      CortexM_v8M.factory ),
    (ARM_ID, CORESIGHT_CLASS, 0xd21, 0x13, 0x4a13) : CmpInfo('ETM',             'M33',      None                ),
    (ARM_ID, CORESIGHT_CLASS, 0xd21, 0x11, 0)      : CmpInfo('TPIU',            'M33',      TPIU.factory        ),
    (ARM_ID, CORESIGHT_CLASS, 0xd22, 0x16, 0x0a06) : CmpInfo('PMU',             'M55',      None                ),
    (ARM_ID, CORESIGHT_CLASS, 0xd22, 0x00, 0x0a07) : CmpInfo('EWIC',            'M55',      None                ),
    (ARM_ID, CORESIGHT_CLASS, 0xd22, 0x43, 0x1a01) : CmpInfo('ITM',             'M55',      ITM.factory         ),
    (ARM_ID, CORESIGHT_CLASS, 0xd22, 0x00, 0x1a02) : CmpInfo('DWT',             'M55',      DWTv2.factory       ),
    (ARM_ID, CORESIGHT_CLASS, 0xd22, 0x00, 0x1a03) : CmpInfo('BPU',             'M55',      FPB.factory         ),
    (ARM_ID, CORESIGHT_CLASS, 0xd22, 0x14, 0x1a14) : CmpInfo('CTI',             'M55',      None                ),
    (ARM_ID, CORESIGHT_CLASS, 0xd22, 0x00, 0x2a04) : CmpInfo('SCS',             'M55',      CortexM_v8M.factory ),
    (ARM_ID, CORESIGHT_CLASS, 0xd22, 0x11, 0)      : CmpInfo('TPIU',            'M55',      TPIU.factory        ),
    (ARM_ID, CORESIGHT_CLASS, 0xd22, 0x13, 0x4a13) : CmpInfo('ETM',             'M55',      None                ),
    (ARM_ID, CORESIGHT_CLASS, 0xd23, 0x16, 0x0a06) : CmpInfo('PMU',             'M85',      None                ),
    (ARM_ID, CORESIGHT_CLASS, 0xd23, 0x00, 0x0a07) : CmpInfo('EWIC',            'M85',      None                ),
    (ARM_ID, CORESIGHT_CLASS, 0xd23, 0x43, 0x1a01) : CmpInfo('ITM',             'M85',      ITM.factory         ),
    (ARM_ID, CORESIGHT_CLASS, 0xd23, 0x00, 0x1a02) : CmpInfo('DWT',             'M85',      DWTv2.factory       ),
    (ARM_ID, CORESIGHT_CLASS, 0xd23, 0x00, 0x1a03) : CmpInfo('BPU',             'M85',      FPB.factory         ),
    (ARM_ID, CORESIGHT_CLASS, 0xd23, 0x14, 0x1a14) : CmpInfo('CTI',             'M85',      None                ),
    (ARM_ID, CORESIGHT_CLASS, 0xd23, 0x00, 0x2a04) : CmpInfo('SCS',             'M85',      CortexM_v8M.factory ),
    (ARM_ID, CORESIGHT_CLASS, 0xd23, 0x11, 0)      : CmpInfo('TPIU',            'M85',      TPIU.factory        ),
    (ARM_ID, CORESIGHT_CLASS, 0xd23, 0x13, 0x4a13) : CmpInfo('ETM',             'M85',      None                ),
    (ARM_ID, CORESIGHT_CLASS, 0xd31, 0x31, 0x0a31) : CmpInfo('MTB',             'M35P',     None                ),
    (ARM_ID, CORESIGHT_CLASS, 0xd31, 0x43, 0x1a01) : CmpInfo('ITM',             'M35P',     ITM.factory         ),
    (ARM_ID, CORESIGHT_CLASS, 0xd31, 0x00, 0x1a02) : CmpInfo('DWT',             'M35P',     DWTv2.factory       ),
    (ARM_ID, CORESIGHT_CLASS, 0xd31, 0x00, 0x1a03) : CmpInfo('BPU',             'M35P',     FPB.factory         ),
    (ARM_ID, CORESIGHT_CLASS, 0xd31, 0x14, 0x1a14) : CmpInfo('CTI',             'M35P',     None                ),
    (ARM_ID, CORESIGHT_CLASS, 0xd31, 0x00, 0x2a04) : CmpInfo('SCS',             'M35P',     CortexM_v8M.factory ),
    (ARM_ID, CORESIGHT_CLASS, 0xd31, 0x13, 0x4a13) : CmpInfo('ETM',             'M35P',     None                ),
    (ARM_ID, CORESIGHT_CLASS, 0xd31, 0x11, 0)      : CmpInfo('TPIU',            'M35P',     TPIU.factory        ),
    (ARM_ID, GENERIC_CLASS,   0x000, 0x00, 0)      : CmpInfo('SCS',             'v7-M',     CortexM.factory     ),
    (ARM_ID, GENERIC_CLASS,   0x001, 0x00, 0)      : CmpInfo('ITM',             'v7-M',     ITM.factory         ),
    (ARM_ID, GENERIC_CLASS,   0x002, 0x00, 0)      : CmpInfo('DWT',             'v7-M',     DWT.factory         ),
    (ARM_ID, GENERIC_CLASS,   0x003, 0x00, 0)      : CmpInfo('FPB',             'v7-M',     FPB.factory         ),
    (ARM_ID, GENERIC_CLASS,   0x004, 0x00, 0)      : CmpInfo('SCS',             'SC300',    CortexM.factory     ),
    (ARM_ID, GENERIC_CLASS,   0x005, 0x00, 0)      : CmpInfo('ITM',             'SC300',    ITM.factory         ),
    (ARM_ID, GENERIC_CLASS,   0x006, 0x00, 0)      : CmpInfo('DWT',             'SC300',    DWT.factory         ),
    (ARM_ID, GENERIC_CLASS,   0x007, 0x00, 0)      : CmpInfo('FPB',             'SC300',    FPB.factory         ),
    (ARM_ID, GENERIC_CLASS,   0x008, 0x00, 0)      : CmpInfo('SCS',             'v6-M',     CortexM.factory     ),
    (ARM_ID, GENERIC_CLASS,   0x00a, 0x00, 0)      : CmpInfo('DWT',             'v6-M',     DWT.factory         ),
    (ARM_ID, GENERIC_CLASS,   0x00b, 0x00, 0)      : CmpInfo('BPU',             'v6-M',     FPB.factory         ),
    (ARM_ID, GENERIC_CLASS,   0x00c, 0x00, 0)      : CmpInfo('SCS',             'v7-M',     CortexM.factory     ),
    (ARM_ID, GENERIC_CLASS,   0x00d, 0x00, 0)      : CmpInfo('SCS',             'SC000',    CortexM.factory     ),
    (ARM_ID, GENERIC_CLASS,   0x00e, 0x00, 0)      : CmpInfo('FPB',             'v7-M',     FPB.factory         ),
    (ARM_ID, SYSTEM_CLASS,    0x101, 0x00, 0)      : CmpInfo('TSGEN',           None,       None                ), # Timestamp Generator
    (ARM_ID, SYSTEM_CLASS,    0x580, 0x00, 0)      : CmpInfo('U55',             None,       None                ), # Ethos U55 NPU
    (FSL_ID, CORESIGHT_CLASS, 0x000, 0x04, 0)      : CmpInfo('MTBDWT',          None,       None                ),
    (STM_ID, SYSTEM_CLASS,    0x000, 0x00, 0)      : CmpInfo('DBGMCU',          None,       None                ),
  # Designer      |Component Class |Part  |Type |Archid           |Name        |Product    |Factory
    (ARM_CHINA_ID, CORESIGHT_CLASS, 0x132, 0x31, 0x0a31) : CmpInfo('MTB',       'Star-MC1', None                ),
    (ARM_CHINA_ID, CORESIGHT_CLASS, 0x132, 0x43, 0x1a01) : CmpInfo('ITM',       'Star-MC1', ITM.factory         ),
    (ARM_CHINA_ID, CORESIGHT_CLASS, 0x132, 0x00, 0x1a02) : CmpInfo('DWT',       'Star-MC1', DWTv2.factory       ),
    (ARM_CHINA_ID, CORESIGHT_CLASS, 0x132, 0x00, 0x1a03) : CmpInfo('BPU',       'Star-MC1', FPB.factory         ),
    (ARM_CHINA_ID, CORESIGHT_CLASS, 0x132, 0x14, 0x1a14) : CmpInfo('CTI',       'Star-MC1', None                ),
    (ARM_CHINA_ID, CORESIGHT_CLASS, 0x132, 0x00, 0x2a04) : CmpInfo('SCS',       'Star-MC1', CortexM_v8M.factory ),
    (ARM_CHINA_ID, CORESIGHT_CLASS, 0x132, 0x13, 0x4a13) : CmpInfo('ETM',       'Star-MC1', None                ),
    (ARM_CHINA_ID, CORESIGHT_CLASS, 0x132, 0x11, 0)      : CmpInfo('TPIU',      'Star-MC1', TPIU.factory        ),
    }




# pyOCD debugger
# Copyright (c) 2021-2022 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
from typing import List
import logging
from pathlib import Path

from .base import SubcommandBase
from ..core.helpers import ConnectHelper
from ..flash.file_programmer import FileProgrammer
from ..utility.cmdline import (
    convert_session_options,
    int_base_0,
)

LOG = logging.getLogger(__name__)

class LoadSubcommand(SubcommandBase):
    """@brief `pyocd load` and `flash` subcommand."""

    NAMES = ['load', 'flash']
    HELP = "Load one or more images into target device memory."
    EPILOG = "Supported file formats are: binary, Intel hex, and ELF32."
    DEFAULT_LOG_LEVEL = logging.WARNING

    ## @brief Valid erase mode options.
    ERASE_OPTIONS = [
        'auto',
        'chip',
        'sector',
        ]

    @classmethod
    def get_args(cls) -> List[argparse.ArgumentParser]:
        """@brief Add this subcommand to the subparsers object."""
        parser = argparse.ArgumentParser(description=cls.HELP, add_help=False)

        parser_options = parser.add_argument_group("load options")
        parser_options.add_argument("-e", "--erase", choices=cls.ERASE_OPTIONS, default='sector',
            help="Choose flash erase method. Default is sector.")
        parser_options.add_argument("-a", "--base-address", metavar="ADDR", type=int_base_0,
            help="Base address used for the address where to write a binary. Defaults to start of flash. "
                 "Only allowed if a single binary file is being loaded.")
        parser_options.add_argument("--trust-crc", action="store_true",
            help="Use only the CRC of each page to determine if it already has the same data.")
        parser_options.add_argument("--format", choices=("bin", "hex", "elf"),
            help="File format. Default is to use the file's extension. If multiple files are provided, then "
                 "all must be of this type.")
        parser_options.add_argument("--skip", metavar="BYTES", default=0, type=int_base_0,
            help="Skip programming the first N bytes. Binary files only.")
        parser_options.add_argument("--no-reset", action="store_true",
            help="Specify to prevent resetting device after programming has finished.")

        parser.add_argument("file", metavar="<file-path>", nargs="+",
            help="File to write to memory. Binary files can have an optional base address appended to the file "
                 "name as '@<address>', for instance 'app.bin@0x20000'.")

        return [cls.CommonOptions.COMMON, cls.CommonOptions.CONNECT, parser]

    def invoke(self) -> int:
        """@brief Handle 'load' subcommand."""
        self._increase_logging(["pyocd.flash.loader", __name__])

        # Validate arguments.
        if (self._args.base_address is not None) and (len(self._args.file) > 1):
            raise ValueError("--base-address cannot be set when loading more than one file; "
                    "use a base address suffix instead")

        session = ConnectHelper.session_with_chosen_probe(
                            project_dir=self._args.project_dir,
                            config_file=self._args.config,
                            user_script=self._args.script,
                            no_config=self._args.no_config,
                            pack=self._args.pack,
                            unique_id=self._args.unique_id,
                            target_override=self._args.target_override,
                            frequency=self._args.frequency,
                            blocking=(not self._args.no_wait),
                            connect_mode=self._args.connect_mode,
                            options=convert_session_options(self._args.options),
                            option_defaults=self._modified_option_defaults(),
                            )
        if session is None:
            LOG.error("No target device available")
            return 1
        with session:
            programmer = FileProgrammer(session,
                            chip_erase=self._args.erase,
                            trust_crc=self._args.trust_crc,
                            no_reset=self._args.no_reset)
            for filename in self._args.file:
                # Get an initial path with the argument as-is.
                file_path = Path(filename).expanduser()

                # Look for a base address suffix. If the supplied argument including an address suffix
                # references an existing file, then the address suffix is not extracted.
                if "@" in filename and not file_path.exists():
                    filename, suffix = filename.rsplit("@", 1)
                    try:
                        base_address = int_base_0(suffix)
                    except ValueError:
                        LOG.error(f'Base address suffix "{suffix}" on file "{filename}" is not a valid integer address')
                        return 1
                else:
                    base_address = self._args.base_address

                # Resolve our path.
                file_path = Path(filename).expanduser().resolve()
                filename = str(file_path)

                if base_address is None:
                    LOG.info("Loading %s", filename)
                else:
                    LOG.info("Loading %s at %#010x", filename, base_address)

                programmer.program(filename,
                                base_address=base_address,
                                skip=self._args.skip,
                                file_format=self._args.format)

        return 0





# pyOCD debugger
# Copyright (c) 2018-2020 Arm Limited
# Copyright (c) 2021-2023 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from time import sleep
import logging
from typing import (Callable, Collection, Dict, List, Optional, overload, Sequence, Set, TYPE_CHECKING, Tuple, Union)
from typing_extensions import (Literal, Protocol)

from .debug_probe import DebugProbe
from ..core import exceptions
from ..core.plugin import Plugin
from ..core.options import OptionInfo
from .pydapaccess import DAPAccess
from ..board.mbed_board import MbedBoard
from ..board.board_ids import (BoardInfo, BOARD_ID_TO_INFO)

if TYPE_CHECKING:
    from types import TracebackType
    from ..board.board import Board

LOG = logging.getLogger(__name__)
TRACE = LOG.getChild("trace")
TRACE.setLevel(logging.CRITICAL)

class _OpenableProtocol(Protocol):
    @property
    def is_open(self) -> bool:
        ...

    def open(self) -> None:
        ...

    def close(self) -> None:
        ...

class _TemporaryOpen:
    """@brief Context manager to ensure the device is open for a short time."""

    def __init__(self, device: _OpenableProtocol, suppress_exceptions: bool = True) -> None:
        self._device = device
        self._suppress_exceptions = suppress_exceptions
        self._did_open_link: bool = False

    def __enter__(self) -> _TemporaryOpen:
        try:
            # Temporarily open the device if not already opened.
            if not self._device.is_open:
                self._device.open()
                self._did_open_link = True
        except (DAPAccess.Error, exceptions.Error) as err:
            if not self._suppress_exceptions:
                raise
            else:
                LOG.debug("suppressing error from attempting to open device %s: %s", self._device, err)

        return self

    def __exit__(self, exc_type: Optional[type], exc_value: Optional[Exception], traceback: Optional[TracebackType]) -> bool:
        # Close the device if we had to open it.
        if self._did_open_link:
            self._device.close()

        # Check for and possibly suppress an exception.
        if (exc_type is not None) and issubclass(exc_type, exceptions.Error):
            if self._suppress_exceptions:
                return True

        return False

class CMSISDAPProbe(DebugProbe):
    """@brief Wraps a pydapaccess link as a DebugProbe.

    Supports CMSIS-DAP v1 and v2.
    """

    # Masks for CMSIS-DAP capabilities.
    SWD_CAPABILITY_MASK = 1
    JTAG_CAPABILITY_MASK = 2

    # Map from DebugProbe protocol types to/from DAPAccess port types.
    #
    # Note that Protocol.DEFAULT gets mapped to PORT.SWD. We need a concrete port type because some
    # non-reference CMSIS-DAP implementations do not accept the default port type.
    _PROTOCOL_TO_PORT: Dict[DebugProbe.Protocol, DAPAccess.PORT] = {
        DebugProbe.Protocol.DEFAULT: DAPAccess.PORT.SWD,
        DebugProbe.Protocol.SWD: DAPAccess.PORT.SWD,
        DebugProbe.Protocol.JTAG: DAPAccess.PORT.JTAG,
        }
    _PORT_TO_PROTOCOL: Dict[DAPAccess.PORT, DebugProbe.Protocol] = {
        DAPAccess.PORT.DEFAULT: DebugProbe.Protocol.DEFAULT,
        DAPAccess.PORT.SWD: DebugProbe.Protocol.SWD,
        DAPAccess.PORT.JTAG: DebugProbe.Protocol.JTAG,
        }

    # APnDP constants.
    DP = 0
    AP = 1

    # Bitmasks for AP register address fields.
    A32 = 0x0000000c

    # Map from AP/DP and 2-bit register address to the enums used by pydapaccess.
    REG_ADDR_TO_ID_MAP: Dict[Tuple[int, int], DAPAccess.REG] = {
        # APnDP A32
        ( 0,    0x0 ) : DAPAccess.REG.DP_0x0,
        ( 0,    0x4 ) : DAPAccess.REG.DP_0x4,
        ( 0,    0x8 ) : DAPAccess.REG.DP_0x8,
        ( 0,    0xC ) : DAPAccess.REG.DP_0xC,
        ( 1,    0x0 ) : DAPAccess.REG.AP_0x0,
        ( 1,    0x4 ) : DAPAccess.REG.AP_0x4,
        ( 1,    0x8 ) : DAPAccess.REG.AP_0x8,
        ( 1,    0xC ) : DAPAccess.REG.AP_0xC,
        }

    ## USB VID and PID pair for DAPLink firmware.
    DAPLINK_VIDPID = (0x0d28, 0x0204)

    @classmethod
    def get_all_connected_probes(
                cls,
                unique_id: Optional[str] = None,
                is_explicit: bool = False
            ) -> Sequence[DebugProbe]:
        try:
            return [cls(dev) for dev in DAPAccess.get_connected_devices()]
        except DAPAccess.Error as exc:
            raise cls._convert_exception(exc) from exc

    @classmethod
    def get_probe_with_id(cls, unique_id: str, is_explicit: bool = False) -> Optional[DebugProbe]:
        try:
            dap_access = DAPAccess.get_device(unique_id)
            if dap_access is not None:
                return cls(dap_access)
            else:
                return None
        except DAPAccess.Error as exc:
            raise cls._convert_exception(exc) from exc

    def __init__(self, device: DAPAccess) -> None:
        super().__init__()
        self._link = device
        self._supported_protocols: List[DebugProbe.Protocol] = []
        self._protocol: Optional[DebugProbe.Protocol] = None
        self._is_open = False
        self._caps: Set[DebugProbe.Capability] = set()

    @property
    def board_id(self) -> Optional[str]:
        """@brief Unique identifier for the board.

        Only board IDs for DAPLink firmware are supported. We can't assume other
        CMSIS-DAP firmware is using the same serial number format, so we cannot reliably
        extract the board ID.

        @return Either a 4-character board ID string, or None if the probe doesn't have a board ID.
        """
        if self._link.vidpid == self.DAPLINK_VIDPID:
            return self.unique_id[0:4]
        else:
            return None

    @property
    def description(self) -> str:
        return self.vendor_name + " " + self.product_name

    @property
    def vendor_name(self) -> str:
        return self._link.vendor_name

    @property
    def product_name(self) -> str:
        return self._link.product_name

    @property
    def supported_wire_protocols(self) -> Collection[DebugProbe.Protocol]:
        """@brief Only valid after opening."""
        return self._supported_protocols

    @property
    def unique_id(self) -> str:
        return self._link.get_unique_id()

    @property
    def wire_protocol(self) -> Optional[DebugProbe.Protocol]:
        return self._protocol

    @property
    def is_open(self) -> bool:
        return self._is_open

    @property
    def capabilities(self) -> Set[DebugProbe.Capability]:
        return self._caps

    @property
    def associated_board_info(self) -> Optional[BoardInfo]:
        """@brief Info about the board associated with this probe, if known."""
        # Get internal board info if available.
        if (self.board_id is not None) and (self.board_id in BOARD_ID_TO_INFO):
            info = BOARD_ID_TO_INFO[self.board_id]
        else:
            info = None

        with _TemporaryOpen(self._link):
            if self._link.supports_board_and_target_names:
                # Get v2.1 board and target info values.
                vendor, board = self._link.board_names
                _, part_number = self._link.target_names

                # Use the target from internal board info in preference, so built-in targets take
                # precedence over DFPs (because the probe will only report part numbers from DFPs).
                if info and info.target:
                    target_device_name = info.target
                elif part_number:
                    target_device_name = part_number.lower().replace("-", "_")
                else:
                    target_device_name = None

                # If we have either target type or board then construct the board info.
                if target_device_name or board:
                    # Vendor can be None, but the BoardInfo must have a valid name.
                    if not board:
                        assert target_device_name
                        board = "Generic " + (part_number or target_device_name)

                    # If we also have the ID based info, then use the test binary from that.
                    binary_name = info.binary if (info is not None) else None

                    # Create a new board info object with the data from the probe.
                    info = BoardInfo(name=board, target=target_device_name, vendor=vendor, binary=binary_name)

        return info

    def create_associated_board(self) -> Optional[Board]:
        assert self.session is not None

        board_info = self.associated_board_info
        if self.board_id or board_info:
            return MbedBoard(self.session, board_info=board_info, board_id=self.board_id)
        return None

    def get_accessible_pins(self, group: DebugProbe.PinGroup) -> Tuple[int, int]:
        """@brief Return masks of pins accessible via the .read_pins()/.write_pins() methods.

        @return Tuple of pin masks for (0) readable, (1) writable pins. See DebugProbe.Pin for mask
        values for those pins that have constants.
        """
        if group is DebugProbe.PinGroup.PROTOCOL_PINS:
            return (self.ProtocolPin.ALL_PINS, self.ProtocolPin.ALL_PINS)
        else:
            return (0, 0)

    def open(self) -> None:
        if self._is_open:
            return
        assert self.session
        try:
            TRACE.debug("trace: open")

            self._link.open()
            self._is_open = True
            self._link.set_deferred_transfer(self.session.options.get('cmsis_dap.deferred_transfers'))

            if self._link.supports_board_and_target_names:
                board_names = self._link.board_names
                target_names = self._link.target_names
                if board_names != (None, None):
                    LOG.debug("Board: %s %s", board_names[0] or "(no vendor)", board_names[1] or "(no name)")
                if target_names != (None, None):
                    LOG.debug("Target: %s %s", target_names[0] or "(no vendor)", target_names[1] or "(no name)")

            # Read CMSIS-DAP capabilities
            caps_value = self._link.identify(DAPAccess.ID.CAPABILITIES)
            if not isinstance(caps_value, int):
                raise exceptions.ProbeError(f"probe {self.unique_id} returned invalid capabilities")
            self._capabilities = caps_value
            self._supported_protocols = [DebugProbe.Protocol.DEFAULT]
            if self._capabilities & self.SWD_CAPABILITY_MASK:
                self._supported_protocols.append(DebugProbe.Protocol.SWD)
            if self._capabilities & self.JTAG_CAPABILITY_MASK:
                self._supported_protocols.append(DebugProbe.Protocol.JTAG)
            # Warn if neither SWD nor JTAG is supported.
            if (self._capabilities & (self.SWD_CAPABILITY_MASK | self.JTAG_CAPABILITY_MASK)) == 0:
                LOG.warning("probe %s reported capabilities indicating it supports neither SWD nor JTAG",
                        self.unique_id)

            self._caps = {
                self.Capability.SWJ_SEQUENCE,
                self.Capability.BANKED_DP_REGISTERS,
                self.Capability.APv2_ADDRESSES,
                self.Capability.JTAG_SEQUENCE,
                self.Capability.PIN_ACCESS,
                }
            if self._link.has_swd_sequence:
                self._caps.add(self.Capability.SWD_SEQUENCE)
            if self._link.has_swo():
                self._caps.add(self.Capability.SWO)
        except DAPAccess.Error as exc:
            raise self._convert_exception(exc) from exc

    def close(self) -> None:
        if not self._is_open:
            return
        try:
            TRACE.debug("trace: close")

            self._link.close()
            self._is_open = False
        except DAPAccess.Error as exc:
            raise self._convert_exception(exc) from exc

    # ------------------------------------------- #
    #          Target control functions
    # ------------------------------------------- #
    def connect(self, protocol: Optional[DebugProbe.Protocol] = None) -> None:
        TRACE.debug("trace: connect(%s)", protocol.name if (protocol is not None) else "None")

        # Convert protocol to port enum.
        #
        # We must get a non-default port, since some CMSIS-DAP implementations do not accept the default
        # port. Note that the conversion of the default port type is contained in the PORT_MAP dict so it
        # is one location.
        port = (self._PROTOCOL_TO_PORT.get(protocol)
                if protocol else self._PROTOCOL_TO_PORT[DebugProbe.Protocol.DEFAULT])
        assert port is not DAPAccess.PORT.DEFAULT

        try:
            self._link.connect(port)
        except DAPAccess.Error as exc:
            raise self._convert_exception(exc) from exc

        # Read the current mode and save it.
        actual_mode = self._link.get_swj_mode()
        assert actual_mode is not None
        self._protocol = self._PORT_TO_PROTOCOL[actual_mode]

    def swj_sequence(self, length: int, bits: int) -> None:
        TRACE.debug("trace: swj_sequence(length=%i, bits=%x)", length, bits)

        try:
            self._link.swj_sequence(length, bits)
        except DAPAccess.Error as exc:
            raise self._convert_exception(exc) from exc

    def swd_sequence(self, sequences: Sequence[Union[Tuple[int], Tuple[int, int]]]) -> Tuple[int, Sequence[bytes]]:
        TRACE.debug("trace: swd_sequence(sequences=%r)", sequences)

        try:
            return self._link.swd_sequence(sequences)
        except DAPAccess.Error as exc:
            raise self._convert_exception(exc) from exc

    def jtag_sequence(self, cycles: int, tms: int, read_tdo: bool, tdi: int) -> Optional[int]:
        TRACE.debug("trace: jtag_sequence(cycles=%i, tms=%x, read_tdo=%s, tdi=%x)", cycles, tms, read_tdo, tdi)

        try:
            self._link.jtag_sequence(cycles, tms, read_tdo, tdi)
        except DAPAccess.Error as exc:
            raise self._convert_exception(exc) from exc

    def disconnect(self) -> None:
        TRACE.debug("trace: disconnect")

        try:
            self._link.disconnect()
            self._protocol = None
        except DAPAccess.Error as exc:
            raise self._convert_exception(exc) from exc

    def set_clock(self, frequency: float) -> None:
        TRACE.debug("trace: set_clock(freq=%i)", frequency)

        try:
            self._link.set_clock(frequency)
        except DAPAccess.Error as exc:
            raise self._convert_exception(exc) from exc

    def reset(self) -> None:
        assert self.session
        TRACE.debug("trace: reset")

        try:
            self._link.assert_reset(True)
            sleep(self.session.options.get('reset.hold_time'))
            self._link.assert_reset(False)
            sleep(self.session.options.get('reset.post_delay'))
        except DAPAccess.Error as exc:
            raise self._convert_exception(exc) from exc

    def assert_reset(self, asserted: bool) -> None:
        TRACE.debug("trace: assert_reset(%s)", asserted)

        try:
            self._link.assert_reset(asserted)
        except DAPAccess.Error as exc:
            raise self._convert_exception(exc) from exc

    def is_reset_asserted(self) -> bool:
        try:
            result = self._link.is_reset_asserted()
            TRACE.debug("trace: is_reset_asserted -> %s", result)
            return result
        except DAPAccess.Error as exc:
            raise self._convert_exception(exc) from exc

    def flush(self) -> None:
        TRACE.debug("trace: flush")

        try:
            self._link.flush()
        except DAPAccess.Error as exc:
            TRACE.debug("trace: error from flush: %r", exc)
            raise self._convert_exception(exc) from exc

    def read_pins(self, group: DebugProbe.PinGroup, mask: int) -> int:
        """@brief Read values of selected debug probe pins.

        See DebugProbe.ProtocolPin for mask values.

        @param self
        @param group Select the pin group to read.
        @param mask Bit mask indicating which pins will be read. The return value will contain only
            bits set in this mask.
        @return Bit mask with the current value of selected pins at each pin's relevant bit position.
        """
        try:
            if group is DebugProbe.PinGroup.PROTOCOL_PINS:
                # CMSIS-DAP DAP_SWJ_Pins command will always return all pin values, so mask
                # the ones the caller wants.
                result = self.from_cmsis_dap_pins(self._link.pin_access(0, 0)) & mask
                TRACE.debug("trace: read_pins(%x) -> %s", mask, result)
                return result
            else:
                return 0
        except DAPAccess.Error as exc:
            raise self._convert_exception(exc) from exc

    def write_pins(self, group: DebugProbe.PinGroup, mask: int, value: int) -> None:
        """@brief Set values of selected debug probe pins.

        See DebugProbe.ProtocolPin for mask values.

        @param self
        @param group Select the pin group to read.
        @param mask Bit mask indicating which pins will be written.
        @param value Mask containing the bit value of to written for selected pins at each pin's
            relevant bit position..
        """
        try:
            if group is DebugProbe.PinGroup.PROTOCOL_PINS:
                self._link.pin_access(self.to_cmsis_dap_pins(mask), self.to_cmsis_dap_pins(value))
                TRACE.debug("trace: write_pins(%s, %s)", mask, value)
        except DAPAccess.Error as exc:
            raise self._convert_exception(exc) from exc

    @staticmethod
    def to_cmsis_dap_pins(mask: int) -> int:
        # - [0] SWCLK/TCK
        # - [1] SWDIO/TMS
        # - [2] TDI
        # - [3] TDO
        # - [5] nTRST
        # - [7] nRESET
        result = 0
        if mask & DebugProbe.ProtocolPin.SWCLK_TCK:
            result |= 1 << 0
        if mask & DebugProbe.ProtocolPin.SWDIO_TMS:
            result |= 1 << 1
        if mask & DebugProbe.ProtocolPin.TDI:
            result |= 1 << 2
        if mask & DebugProbe.ProtocolPin.TDO:
            result |= 1 << 3
        if mask & DebugProbe.ProtocolPin.nRESET:
            result |= 1 << 7
        if mask & DebugProbe.ProtocolPin.nTRST:
            result |= 1 << 5
        return result

    @staticmethod
    def from_cmsis_dap_pins(mask: int) -> int:
        result = 0
        if mask & (1 << 0):
            result |= DebugProbe.ProtocolPin.SWCLK_TCK
        if mask & (1 << 1):
            result |= DebugProbe.ProtocolPin.SWDIO_TMS
        if mask & (1 << 2):
            result |= DebugProbe.ProtocolPin.TDI
        if mask & (1 << 3):
            result |= DebugProbe.ProtocolPin.TDO
        if mask & (1 << 5):
            result |= DebugProbe.ProtocolPin.nTRST
        if mask & (1 << 7):
            result |= DebugProbe.ProtocolPin.nRESET
        return result

    # ------------------------------------------- #
    #          DAP Access functions
    # ------------------------------------------- #

    @overload
    def read_dp(self, addr: int) -> int:
        ...

    @overload
    def read_dp(self, addr: int, now: Literal[True] = True) -> int:
        ...

    @overload
    def read_dp(self, addr: int, now: Literal[False]) -> Callable[[], int]:
        ...

    @overload
    def read_dp(self, addr: int, now: bool) -> Union[int, Callable[[], int]]:
        ...

    def read_dp(self, addr: int, now: bool = True) -> Union[int, Callable[[], int]]:
        reg_id = self.REG_ADDR_TO_ID_MAP[self.DP, addr]

        try:
            if not now:
                TRACE.debug("trace: read_dp(addr=%#010x) -> ...", addr)
            result = self._link.read_reg(reg_id, now=now)
        except DAPAccess.Error as error:
            TRACE.debug("trace: read_dp(addr=%#010x) -> error(%s)", addr, error)
            raise self._convert_exception(error) from error

        # Read callback returned for async reads.
        def read_dp_result_callback():
            try:
                value = result()
                TRACE.debug("trace: ... read_dp(addr=%#010x) -> %#010x", addr, value)
                return value
            except DAPAccess.Error as error:
                TRACE.debug("trace: ... read_dp(addr=%#010x) -> error(%s)", addr, error)
                raise self._convert_exception(error) from error

        if now:
            TRACE.debug("trace: read_dp(addr=%#010x) -> %#010x", addr, result)
            return result
        else:
            return read_dp_result_callback

    def write_dp(self, addr: int, data: int) -> None:
        reg_id = self.REG_ADDR_TO_ID_MAP[self.DP, addr]

        # Write the DP register.
        try:
            self._link.write_reg(reg_id, data)
            TRACE.debug("trace: write_dp(addr=%#010x, data=%#010x)", addr, data)
        except DAPAccess.Error as error:
            TRACE.debug("trace: write_dp(addr=%#010x, data=%#010x) -> error(%s)", addr, data, error)
            raise self._convert_exception(error) from error

    @overload
    def read_ap(self, addr: int) -> int:
        ...

    @overload
    def read_ap(self, addr: int, now: Literal[True] = True) -> int:
        ...

    @overload
    def read_ap(self, addr: int, now: Literal[False]) -> Callable[[], int]:
        ...

    @overload
    def read_ap(self, addr: int, now: bool) -> Union[int, Callable[[], int]]:
        ...

    def read_ap(self, addr: int, now: bool = True) -> Union[int, Callable[[], int]]:
        assert isinstance(addr, int)
        ap_reg = self.REG_ADDR_TO_ID_MAP[self.AP, (addr & self.A32)]

        try:
            if not now:
                TRACE.debug("trace: read_ap(addr=%#010x) -> ...", addr)
            result = self._link.read_reg(ap_reg, now=now)
        except DAPAccess.Error as error:
            raise self._convert_exception(error) from error

        # Read callback returned for async reads.
        def read_ap_result_callback():
            try:
                value = result()
                TRACE.debug("trace: ... read_ap(addr=%#010x) -> %#010x", addr, value)
                return value
            except DAPAccess.Error as error:
                TRACE.debug("trace: ... read_ap(addr=%#010x) -> error(%s)", addr, error)
                raise self._convert_exception(error) from error

        if now:
            TRACE.debug("trace: read_ap(addr=%#010x) -> %#010x", addr, result)
            return result
        else:
            return read_ap_result_callback

    def write_ap(self, addr: int, data) -> None:
        assert isinstance(addr, int)
        ap_reg = self.REG_ADDR_TO_ID_MAP[self.AP, (addr & self.A32)]

        try:
            # Perform the AP register write.
            self._link.write_reg(ap_reg, data)
            TRACE.debug("trace: write_ap(addr=%#010x, data=%#010x)", addr, data)
        except DAPAccess.Error as error:
            TRACE.debug("trace: write_ap(addr=%#010x, data=%#010x) -> error(%s)", addr, data, error)
            raise self._convert_exception(error) from error

    @overload
    def read_ap_multiple(self, addr: int, count: int = 1) -> Sequence[int]:
        ...

    @overload
    def read_ap_multiple(self, addr: int, count: int, now: Literal[True] = True) -> Sequence[int]:
        ...

    @overload
    def read_ap_multiple(self, addr: int, count: int, now: Literal[False]) -> Callable[[], Sequence[int]]:
        ...

    @overload
    def read_ap_multiple(self, addr: int, count: int, now: bool) -> Union[Sequence[int], Callable[[], Sequence[int]]]:
        ...

    def read_ap_multiple(self, addr: int, count: int = 1, now: bool = True) \
             -> Union[Sequence[int], Callable[[], Sequence[int]]]:
        assert isinstance(addr, int)
        ap_reg = self.REG_ADDR_TO_ID_MAP[self.AP, (addr & self.A32)]

        try:
            if not now:
                TRACE.debug("trace: read_ap_multi(addr=%#010x, count=%i) -> ...", addr, count)
            result = self._link.reg_read_repeat(count, ap_reg, dap_index=0, now=now)
        except DAPAccess.Error as exc:
            raise self._convert_exception(exc) from exc

        # Need to wrap the deferred callback to convert exceptions.
        def read_ap_repeat_callback():
            try:
                values = result()
                TRACE.debug("trace: ... read_ap_multi(addr=%#010x, count=%i) -> [%s]", addr, count,
                        ", ".join(["%#010x" % v for v in values]))
                return values
            except DAPAccess.Error as exc:
                TRACE.debug("trace: ... read_ap_multi(addr=%#010x, count=%i) -> error(%s)",
                    addr, count, exc)
                raise self._convert_exception(exc) from exc

        if now:
            TRACE.debug("trace: read_ap_multi(addr=%#010x, count=%i) -> [%s]", addr, count,
                    ", ".join(["%#010x" % v for v in result])) # type: ignore # result is always iterable if now is True
            return result
        else:
            return read_ap_repeat_callback

    def write_ap_multiple(self, addr: int, values) -> None:
        assert isinstance(addr, int)
        ap_reg = self.REG_ADDR_TO_ID_MAP[self.AP, (addr & self.A32)]

        try:
            self._link.reg_write_repeat(len(values), ap_reg, values, dap_index=0)
            TRACE.debug("trace: write_ap_multi(addr=%#010x, (%i)[%s])", addr, len(values),
                   ", ".join(["%#010x" % v for v in values]))
        except DAPAccess.Error as exc:
            TRACE.debug("trace: write_ap_multi(addr=%#010x, (%i)[%s]) -> error(%s)", addr, len(values),
                    ", ".join(["%#010x" % v for v in values]), exc)
            raise self._convert_exception(exc) from exc

    # ------------------------------------------- #
    #          SWO functions
    # ------------------------------------------- #

    def swo_start(self, baudrate: float) -> None:
        TRACE.debug("trace: swo_start(baud=%i)", baudrate)

        try:
            self._link.swo_configure(True, baudrate)
            self._link.swo_control(True)
        except DAPAccess.Error as exc:
            raise self._convert_exception(exc) from exc

    def swo_stop(self) -> None:
        TRACE.debug("trace: swo_stop")

        try:
            self._link.swo_configure(False, 0)
        except DAPAccess.Error as exc:
            raise self._convert_exception(exc) from exc

    def swo_read(self) -> bytearray:
        try:
            data = self._link.swo_read()
            TRACE.debug("trace: swo_read -> %i bytes", len(data))
            return data
        except DAPAccess.Error as exc:
            raise self._convert_exception(exc) from exc

    @staticmethod
    def _convert_exception(exc: Exception) -> Exception:
        if isinstance(exc, DAPAccess.TransferFaultError):
            return exceptions.TransferFaultError(*exc.args)
        elif isinstance(exc, DAPAccess.TransferTimeoutError):
            return exceptions.TransferTimeoutError(*exc.args)
        elif isinstance(exc, DAPAccess.TransferError):
            return exceptions.TransferError(*exc.args)
        elif isinstance(exc, (DAPAccess.DeviceError, DAPAccess.CommandError)):
            return exceptions.ProbeError(*exc.args)
        elif isinstance(exc, DAPAccess.Error):
            return exceptions.Error(*exc.args)
        else:
            return exc

class CMSISDAPProbePlugin(Plugin):
    """@brief Plugin class for CMSISDAPProbe."""

    def load(self):
        return CMSISDAPProbe

    @property
    def name(self):
        return "cmsisdap"

    @property
    def description(self):
        return "CMSIS-DAP debug probe"

    @property
    def options(self):
        """@brief Returns CMSIS-DAP probe options."""
        return [
            OptionInfo('cmsis_dap.deferred_transfers', bool, True,
                "Whether the CMSIS-DAP probe backend will use deferred transfers for improved performance."),
            OptionInfo('cmsis_dap.limit_packets', bool, False,
                "Restrict CMSIS-DAP backend to using a single in-flight command at a time."),
            ]



# pyOCD debugger
# Copyright (c) 2020 Arm Limited
# Copyright (c) 2021 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from .debug_probe import DebugProbe
from ..core import exceptions

LOG = logging.getLogger(__name__)

class SharedDebugProbeProxy(object):
    """@brief Proxy for a DebugProbe that allows it to be shared by multiple clients.

    The main purpose of this class is to keep track of the number of times the probe has been
    opened and connected, and to perform checks to ensure that probes don't interfere with each
    other. Most probe APIs are simply passed to the underlying probe object.
    """

    def __init__(self, probe):
        self._session = None
        self._probe = probe
        self._open_count = 0
        self._connect_count = 0

    @property
    def session(self):
        """@brief Session associated with this probe."""
        return self._session

    @session.setter
    def session(self, the_session):
        self._session = the_session
        self._probe.session = the_session

    @property
    def probe(self):
        return self._probe

    def open(self):
        if self._open_count == 0:
            self._probe.open()
        self._open_count += 1

    def close(self):
        if self._open_count == 1:
            self._probe.close()
        self._open_count -= 1

    def connect(self, protocol=None):
        # First to connect gets to choose the protocol.
        if self._connect_count == 0:
            self._probe.connect(protocol)
        elif protocol not in (DebugProbe.Protocol.DEFAULT, self._probe.wire_protocol):
            raise exceptions.ProbeError("probe already connected using %s protocol" % self._probe.wire_protocol.name)
        self._connect_count += 1

    def disconnect(self):
        if self._connect_count == 1:
            self._probe.disconnect()
        self._connect_count -= 1

    def swj_sequence(self, length, bits):
        self._probe.swj_sequence(length, bits)

    def __getattr__(self, name):
        """@brief Redirect to underlying probe object methods."""
        if hasattr(self._probe, name):
            return getattr(self._probe, name)
        else:
            raise AttributeError(name)




# pyOCD debugger
# Copyright (c) 2017-2019 Arm Limited
# Copyright (c) 2021-2022 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import struct
import logging
import itertools
from typing import (TYPE_CHECKING, Any, Dict, Iterator, List, Optional, Sequence, Set, Tuple)

from ...debug.elf.elf import ELFBinaryFile
from ...utility.compatibility import to_str_safe
from ...core.memory_map import MemoryRange
from ...core import exceptions
from ...utility.conversion import byte_list_to_u32le_list
from ...utility.mask import align_down

if TYPE_CHECKING:
    from ...debug.elf.elf import ELFSection
    from ...core.memory_map import RamRegion

LOG = logging.getLogger(__name__)

class FlashAlgoException(exceptions.TargetSupportError):
    """@brief Exception class for errors parsing an FLM file."""
    pass

RoRwZiType = Tuple[Optional["ELFSection"], Optional["ELFSection"], Optional[MemoryRange]]

class PackFlashAlgo:
    """@brief Class to wrap a flash algo

    This class is intended to provide easy access to the information
    provided by a flash algorithm, such as symbols and the flash
    algorithm itself.

    @sa PackFlashInfo
    """

    REQUIRED_SYMBOLS = {
        "Init",
        "UnInit",
        "EraseSector",
        "ProgramPage",
        }

    EXTRA_SYMBOLS = {
        "BlankCheck",
        "EraseChip",
        "Verify",
        }

    SECTIONS_TO_FIND = (
        ("PrgCode", "SHT_PROGBITS"),
        ("PrgData", "SHT_PROGBITS"),
        ("PrgData", "SHT_NOBITS"),
        )

    ## @brief Standard flash blob header with a breakpoint instruction.
    #
    # This header consists of two instructions:
    #
    # ```
    # bkpt  #0
    # b     .-2     # branch to the bkpt
    # ```
    #
    # Before running a flash algo operation, LR is set to the address of the `bkpt` instruction,
    # so when the operation function returns it will halt the CPU.
    _FLASH_BLOB_HEADER = [ 0xE7FDBE00 ]
    ## @brief Size of the flash blob header in bytes.
    _FLASH_BLOB_HEADER_SIZE = len(_FLASH_BLOB_HEADER) * 4

    # Minimum size that must be allocated for the flash algo stack.
    _MIN_STACK_SIZE = 512

    # Alignment for page buffers.
    _PAGE_BUFFER_ALIGN = 16

    def __init__(self, data):
        """@brief Construct a PackFlashAlgo from a file-like object."""
        self.elf = ELFBinaryFile(data)
        self.flash_info = PackFlashInfo(self.elf)

        self.flash_start = self.flash_info.start
        self.flash_size = self.flash_info.size
        self.page_size = self.flash_info.page_size
        self.sector_sizes = self.flash_info.sector_info_list

        symbols: Dict[str, int] = {}
        x = self._extract_symbols(self.REQUIRED_SYMBOLS)
        symbols.update(x)
        symbols.update(self._extract_symbols(self.EXTRA_SYMBOLS,
                                        default=0xFFFFFFFF))
        self.symbols = symbols

        ro_rw_zi = self._find_sections(self.SECTIONS_TO_FIND)
        ro_rw_zi = self._algo_fill_zi_if_missing(ro_rw_zi)
        error_msg = self._algo_check_for_section_problems(ro_rw_zi)
        if error_msg is not None:
            raise FlashAlgoException(error_msg)

        sect_ro, sect_rw, sect_zi = ro_rw_zi
        assert sect_ro and sect_rw and sect_zi
        self.ro_start = sect_ro.start
        self.ro_size = sect_ro.length
        self.rw_start = sect_rw.start
        self.rw_size = sect_rw.length
        self.zi_start = sect_zi.start
        self.zi_size = sect_zi.length

        self.algo_data = self._create_algo_bin(ro_rw_zi)

    def iter_sector_size_ranges(self) -> Iterator[Tuple[MemoryRange, int]]:
        """@brief Iterator yielding tuples with a memory ranges and sector size for each of the algo's
            sector sizes.
        """
        # The sector_sizes attribute is a list of bi-tuples of (start-address, sector-size), sorted by start address.
        for j, (offset, sector_size) in enumerate(self.sector_sizes):
            start = self.flash_start + offset

            # Determine the end address of the this sector range. For the last range, the end
            # is just the end of the entire region. Otherwise it's the start of the next
            # range - 1.
            if j + 1 >= len(self.sector_sizes):
                end = self.flash_start + self.flash_size - 1
            else:
                end = self.flash_start + self.sector_sizes[j + 1][0] - 1

            # Skip wrong start and end addresses
            if end < start:
                continue

            yield MemoryRange(start, end), sector_size

    def get_pyocd_flash_algo(self, blocksize: int, ram_region: "RamRegion") -> Dict[str, Any]:
        """@brief Return a dictionary representing a pyOCD flash algorithm, or None.

        The most interesting operation this method performs is dynamically allocating memory
        for the flash algo from a given RAM region. Note that the .data and .bss sections are
        concatenated with .text. That's why there isn't a specific allocation for those sections.

        Double buffering is supported as long as there is enough RAM.

        Memory layout:
        ```
        [<--stack] [buf2] [buf1] [code]
        ^ ram start                   ^ ram end
        ```

        @param self
        @param blocksize The size to use for page buffers, normally the erase block size.
        @param ram_region A RamRegion object where the flash algo will be allocated.
        @return A pyOCD-style flash algo dictionary. If None is returned, the flash algo did
            not fit into the provided ram_region.

        @exception FlashAlgoException Raised if the algo cannot be placed in memory.
        """
        instructions = self._FLASH_BLOB_HEADER + byte_list_to_u32le_list(self.algo_data)

        # Start at end of RAM and work backwards.
        addr = ram_region.start + ram_region.length

        # Load address
        addr -= len(instructions) * 4
        addr_load = addr

        # Data buffer 1
        unaligned_buffer_addr = addr - blocksize
        addr = align_down(unaligned_buffer_addr, self._PAGE_BUFFER_ALIGN)
        addr_data = addr

        if addr_data < ram_region.start:
            # Not enough space for flash algorithm
            raise FlashAlgoException("not enough memory space to fit flash algorithm")

        # Data buffer 2
        unaligned_buffer_addr = addr - blocksize
        addr = align_down(unaligned_buffer_addr, self._PAGE_BUFFER_ALIGN)
        addr_data2 = addr

        # Stack
        # Select best fit for one or two data buffers and a variable size stack.
        # TODO Switching down from two to one buffer should probably be done with the stack size around
        #   mid-level instead of going all the way down to minimum first.
        stack_size_one_buf = addr_data - ram_region.start
        stack_size_two_bufs = addr_data2 - ram_region.start

        if stack_size_two_bufs < self._MIN_STACK_SIZE:
            # One buffer
            stack_size = stack_size_one_buf
            addr_stack = addr_data
            page_buffers = [addr_data]

            LOG.debug("flash algo: [stack=%#x; %#x b] [b1=%#x,+%#x] [code=%#x,+%#x,%#x b] (ram=%#010x, %#x b)",
                addr_stack, stack_size,
                addr_data, addr_data - ram_region.start,
                addr_load, addr_load - ram_region.start, len(instructions) * 4,
                ram_region.start, ram_region.length
            )
        else:
            stack_size = stack_size_two_bufs
            addr_stack = addr_data2
            page_buffers = [addr_data, addr_data2]

            LOG.debug("flash algo: [stack=%#x; %#x b] [b2=%#x,+%#x] [b1=%#x,+%#x] [code=%#x,+%#x,%#x b] (ram=%#010x, %#x b)",
                addr_stack, stack_size,
                addr_data, addr_data - ram_region.start,
                addr_data2, addr_data2 - ram_region.start,
                addr_load, addr_load - ram_region.start, len(instructions) * 4,
                ram_region.start, ram_region.length
            )
        # TODO - analyzer support

        code_start = addr_load + self._FLASH_BLOB_HEADER_SIZE
        flash_algo = {
            "load_address": addr_load,
            "instructions": instructions,
            "pc_init": code_start + self.symbols["Init"],
            "pc_unInit": code_start + self.symbols["UnInit"],
            "pc_eraseAll": code_start + self.symbols["EraseChip"],
            "pc_erase_sector": code_start + self.symbols["EraseSector"],
            "pc_program_page": code_start + self.symbols["ProgramPage"],
            "page_buffers": page_buffers,
            "begin_data": page_buffers[0],
            "begin_stack": addr_stack,
            "end_stack": addr_stack - stack_size,
            "static_base": code_start + self.rw_start,
            "min_program_length": self.page_size,
            "analyzer_supported": False
        }
        return flash_algo

    def _extract_symbols(self, symbols: Set[str], default: Optional[int] = None) -> Dict[str, int]:
        """@brief Fill 'symbols' field with required flash algo symbols"""
        to_ret: Dict[str, int] = {}
        for symbol in symbols:
            symbolInfo = self.elf.symbol_decoder.get_symbol_for_name(symbol)
            if symbolInfo is None:
                if default is not None:
                    to_ret[symbol] = default
                    continue
                raise FlashAlgoException("Missing symbol %s" % symbol)
            to_ret[symbol] = symbolInfo.address
        return to_ret

    def _find_sections(self, name_type_pairs: Sequence[Tuple[str, str]]) -> Tuple[Optional["ELFSection"], ...]:
        """@brief Return a list of sections the same length and order of the input list"""
        sections: List[Optional["ELFSection"]] = [None] * len(name_type_pairs)
        for section in self.elf.sections:
            section_name = to_str_safe(section.name)
            section_type = section.type
            for i, name_and_type in enumerate(name_type_pairs):
                if name_and_type != (section_name, section_type):
                    continue
                if sections[i] is not None:
                    raise FlashAlgoException("Elf contains duplicate section %s attr %s" %
                                    (section_name, section_type))
                sections[i] = section
        return tuple(sections)

    def _algo_fill_zi_if_missing(self, ro_rw_zi: RoRwZiType) -> RoRwZiType:
        """@brief Create an empty zi section if it is missing"""
        s_ro, s_rw, s_zi = ro_rw_zi
        if s_rw is None:
            return ro_rw_zi
        if s_zi is not None:
            return ro_rw_zi
        s_zi = MemoryRange(start=(s_rw.start + s_rw.length), length=0)
        return s_ro, s_rw, s_zi

    def _algo_check_for_section_problems(self, ro_rw_zi: RoRwZiType) -> Optional[str]:
        """@brief Return a string describing any errors with the layout or None if good"""
        s_ro, s_rw, s_zi = ro_rw_zi
        if s_ro is None:
            return "RO section is missing"
        if s_rw is None:
            return "RW section is missing"
        if s_zi is None:
            return "ZI section is missing"
        if s_ro.start != 0:
            return "RO section does not start at address 0"
        if s_ro.start + s_ro.length != s_rw.start:
            return "RW section does not follow RO section"
        if s_rw.start + s_rw.length != s_zi.start:
            return "ZI section does not follow RW section"
        return None

    def _create_algo_bin(self, ro_rw_zi: RoRwZiType):
        """Create a binary blob of the flash algo which can execute from ram"""
        sect_ro, sect_rw, sect_zi = ro_rw_zi
        assert sect_ro and sect_rw and sect_zi
        algo_size = sect_ro.length + sect_rw.length + sect_zi.length
        algo_data = bytearray(algo_size)
        for section in (sect_ro, sect_rw):
            start = section.start
            size = section.length
            data = section.data
            assert len(data) == size
            algo_data[start:start + size] = data
        return algo_data


class PackFlashInfo(object):
    """@brief Wrapper class for the non-executable information in an FLM file"""

    FLASH_DEVICE_STRUCT = "<H128sHLLLLBxxxLL"
    FLASH_DEVICE_STRUCT_SIZE = struct.calcsize(FLASH_DEVICE_STRUCT)
    FLASH_SECTORS_STRUCT = "<LL"
    FLASH_SECTORS_STRUCT_SIZE = struct.calcsize(FLASH_SECTORS_STRUCT)
    SECTOR_END = 0xFFFFFFFF

    def __init__(self, elf):
        dev_info = elf.symbol_decoder.get_symbol_for_name("FlashDevice")
        if dev_info is None:
            values: Sequence[Any] = [0] * 10
            values[1] = b""
            self.sector_info_list = []
            info_start = 0
            info_size = 0
        else:
            info_start = dev_info.address
            info_size = struct.calcsize(self.FLASH_DEVICE_STRUCT)
            data = elf.read(info_start, self.FLASH_DEVICE_STRUCT_SIZE)
            values = struct.unpack(self.FLASH_DEVICE_STRUCT, data)

        self.version = values[0]
        self.name = values[1].strip(b"\x00")
        self.type = values[2]
        self.start = values[3]
        self.size = values[4]
        self.page_size = values[5]
        self.value_empty = values[7]
        self.prog_timeout_ms = values[8]
        self.erase_timeout_ms = values[9]

        if dev_info is not None:
            sector_gen = self._sector_and_sz_itr(elf, info_start + info_size)
            self.sector_info_list = list(sector_gen)

    def __str__(self):
        desc =  "Flash Device:" + os.linesep
        desc += "  name=%s" % self.name + os.linesep
        desc += "  version=0x%x" % self.version + os.linesep
        desc += "  type=%i" % self.type + os.linesep
        desc += "  start=0x%x" % self.start + os.linesep
        desc += "  size=0x%x" % self.size + os.linesep
        desc += "  page_size=0x%x" % self.page_size + os.linesep
        desc += "  value_empty=0x%x" % self.value_empty + os.linesep
        desc += "  prog_timeout_ms=%i" % self.prog_timeout_ms + os.linesep
        desc += "  erase_timeout_ms=%i" % self.erase_timeout_ms + os.linesep
        desc += "  sectors:" + os.linesep
        for sector_start, sector_size in self.sector_info_list:
            desc += ("    start=0x%x, size=0x%x" %
                     (sector_start, sector_size) + os.linesep)
        return desc

    def _sector_and_sz_itr(self, elf, data_start):
        """Iterator which returns starting address and sector size"""
        for entry_start in itertools.count(data_start, self.FLASH_SECTORS_STRUCT_SIZE):
            data = elf.read(entry_start, self.FLASH_SECTORS_STRUCT_SIZE)
            size, start = struct.unpack(self.FLASH_SECTORS_STRUCT, data)
            start_and_size = start, size
            if start_and_size == (self.SECTOR_END, self.SECTOR_END):
                return
            yield start_and_size





# pyOCD debugger
# Copyright (c) 2019 Arm Limited
# COpyright (c) 2021-2022 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import (Any, Optional, TYPE_CHECKING)

if TYPE_CHECKING:
    from .session import Session
    from .soc_target import SoCTarget
    from .core_target import CoreTarget
    from .target import Target
    from ..board.board import Board
    from ..utility.sequencer import CallSequence
    from ..debug.sequences.delegates import DebugSequenceDelegate
    from ..commands.execution_context import CommandSet

## @brief Return type for some delegate methods.
#
# For certain delegate method, the delegate can reply with a boolean or None, where True means that
# it handled the actions and no further action is to be performed, and False or None means to continue
# processing.
DelegateResult = Optional[bool]


class DelegateHavingMixIn:
    """! @brief Mix-in class adding delegate support.

    There are two types of delegates:
    - Debug sequence delegates
    - Arbitrary delegate classes
    """

    _delegate: Any = None
    _debug_sequence_delegate: Optional["DebugSequenceDelegate"] = None

    @property
    def delegate(self) -> Any:
        return self._delegate

    @delegate.setter
    def delegate(self, the_delegate: Any) -> None:
        self._delegate = the_delegate

    @property
    def debug_sequence_delegate(self) -> Optional["DebugSequenceDelegate"]:
        return self._debug_sequence_delegate

    @debug_sequence_delegate.setter
    def debug_sequence_delegate(self, the_delegate: "DebugSequenceDelegate") -> None:
        self._debug_sequence_delegate = the_delegate

    def delegate_implements(self, method_name: str) -> bool:
        return (self._delegate is not None) and (hasattr(self._delegate, method_name))

    def call_delegate(self, method_name: str, *args: Any, **kwargs: Any) -> Optional[bool]:
        if self.delegate_implements(method_name):
            return getattr(self._delegate, method_name)(*args, **kwargs)
        else:
            # The default action is always taken if None is returned.
            return None

    def has_debug_sequence(self, name: str, pname: Optional[str] = None) -> bool:
        seq_delegate = self.debug_sequence_delegate
        return seq_delegate.has_sequence_with_name(name, pname) if seq_delegate else False


class TargetDelegateInterface:
    """@brief Abstract class defining the delegate interface for targets.

    Note that delegates don't actually have to derive from this class due to Python's
    dynamic method dispatching. The primary purpose of this class is for documentation
    and type checking.
    """

    def __init__(self, session: "Session") -> None:
        self._session = session

    def unlock_device(self, target: "SoCTarget") -> None:
        """! @brief Hook to perform any required unlock sequence.

        Called after the DP is initialised but prior to discovery.

        @param self
        @param target An SoCTarget object about to be initialized.
        @return Ignored.
        """
        pass

    def will_connect(self, board: "Board") -> None:
        """@brief Pre-init hook for the board.
        @param self
        @param board A Board instance that is about to be initialized.
        @return Ignored.
        """
        pass

    def did_connect(self, board: "Board") -> None:
        """@brief Post-initialization hook for the board.
        @param self
        @param board A Board instance.
        @return Ignored.
        """
        pass

    def will_init_target(self, target: "SoCTarget", init_sequence: "CallSequence") -> None:
        """@brief Hook to review and modify init call sequence prior to execution.
        @param self
        @param target An SoCTarget object about to be initialized.
        @param init_sequence The CallSequence that will be invoked. Because call sequences are
            mutable, this parameter can be modified before return to change the init calls.
        @return Ignored.
        """
        pass

    def did_init_target(self, target: "SoCTarget") -> None:
        """@brief Post-initialization hook.
        @param self
        @param target An SoCTarget.
        @return Ignored.
        """
        pass

    def will_start_debug_core(self, core: "CoreTarget") -> None:
        """@brief Notification hook for before core debug is enabled.

        This hook is called during connection, prior to any register accesses being performed on the
        indicated core (aside from the CoreSight peripheral ID registers that were read to identify
        the core's presence during discovery).

        @param self
        @param core A CoreTarget object about to be initialized.
        @return Ignored.
        """
        pass

    def start_debug_core(self, core: "CoreTarget") -> DelegateResult:
        """@brief Core debug initialization hook.
        @param self
        @param core A CoreTarget object.
        @retval True Do not perform the normal procedure to start core debug.
        @retval "False or None" Continue with normal behaviour.
        """
        pass

    def did_start_debug_core(self, core: "CoreTarget") -> None:
        """@brief Notification hook that core debug has been enabled.

        This hook method is called once a debug has been enabled for a core, and it has been fully
        identified.

        @param self
        @param core A CoreTarget object.
        @return Ignored.
        """
        pass

    def will_stop_debug_core(self, core: "CoreTarget") -> None:
        """@brief Pre core disconnect notification hook for the core.
        @param self
        @param core A CoreTarget object.
        @return Ignored.
        """
        pass

    def stop_debug_core(self, core: "CoreTarget") -> DelegateResult:
        """@brief Core debug disable hook.
        @param self
        @param core A CoreTarget object.
        @retval True Do not perform the normal procedure to disable core debug.
        @retval "False or None" Continue with normal behaviour.
        """
        pass

    def did_stop_debug_core(self, core: "CoreTarget") -> None:
        """@brief Post core disconnect notification hook for the core.
        @param self
        @param core A CoreTarget object.
        @return Ignored.
        """
        pass

    def will_disconnect(self, target: "SoCTarget", resume: bool) -> None:
        """@brief Pre-disconnect hook.
        @param self
        @param target Either a CoreSightTarget or CortexM object.
        @param resume The value of the `disconnect_on_resume` option.
        @return Ignored.
        """
        pass

    def did_disconnect(self, target: "SoCTarget", resume: bool) -> None:
        """@brief Post-disconnect hook.
        @param self
        @param target Either a CoreSightTarget or CortexM object.
        @param resume The value of the `disconnect_on_resume` option.
        @return Ignored."""
        pass

    def will_reset(self, core: "Target", reset_type: "Target.ResetType") -> DelegateResult:
        """@brief Pre-reset hook.
        @param self
        @param core A CortexM instance.
        @param reset_type One of the Target.ResetType enumerations.
        @retval True
        @retval "False or None"
        """
        pass

    def did_reset(self, core: "Target", reset_type: "Target.ResetType") -> None:
        """@brief Post-reset hook.
        @param self
        @param core A CortexM instance.
        @param reset_type One of the Target.ResetType enumerations.
        @return Ignored.
        """
        pass

    def set_reset_catch(self, core: "CoreTarget", reset_type: "Target.ResetType") -> DelegateResult:
        """@brief Hook to prepare target for halting on reset.
        @param self
        @param core A CortexM instance.
        @param reset_type One of the Target.ResetType enumerations.
        @retval True
        @retval "False or None"
        """
        pass

    def clear_reset_catch(self, core: "CoreTarget", reset_type: "Target.ResetType") -> None:
        """@brief Hook to clean up target after a reset and halt.
        @param self
        @param core A CortexM instance.
        @param reset_type
        @return Ignored.
        """
        pass

    def mass_erase(self, target: "SoCTarget") -> DelegateResult:
        """@brief Hook to override mass erase.
        @param self
        @param target A CoreSightTarget object.
        @retval True Indicate that mass erase was performed by the hook.
        @retval "False or None" Mass erase was not overridden and the caller should proceed with the standard
            mass erase procedure.
        """
        pass

    def trace_start(self, target: "SoCTarget", mode: int) -> None:
        """@brief Hook to prepare for tracing the target.
        @param self
        @param target A CoreSightTarget object.
        @param mode The trace mode. Currently always 0 to indicate SWO.
        @return Ignored.
        """
        pass

    def trace_stop(self, target: "SoCTarget", mode: int) -> None:
        """@brief Hook to clean up after tracing the target.
        @param self
        @param target A CoreSightTarget object.
        @param mode The trace mode. Currently always 0 to indicate SWO.
        @return Ignored.
        """
        pass

    def add_target_command_groups(self, target: "SoCTarget", command_set: "CommandSet"):
        """@brief Hook for adding target-specific commands to a command set.
        @param self
        @param target A `SoCTarget` object.
        @param command_set The `CommandSet` object to which commands may be added.
        """
        pass




# pyOCD debugger
# Copyright (c) 2020 Arm Limited
# Copyright (c) 2021-2023 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from importlib_metadata import entry_points
from typing import (
    Any,
    Dict,
    List,
    )

from .._version import version as pyocd_version
from .options import (
    add_option_set,
    OptionInfo,
    )

LOG = logging.getLogger(__name__)

class Plugin:
    """@brief Class that describes a plugin for pyOCD.

    Each plugin vends a subclass of Plugin that describes itself and provides meta-actions.

    An instance is created and queried for whether the plugin can be loaded by calling
    should_load(). If this method returns True, then load() is called. The default implementation
    will always load, and does nothing when loaded.
    """

    def should_load(self) -> bool:
        """@brief Whether the plugin should be loaded."""
        return True

    def load(self) -> Any:
        """@brief Load the plugin and return the plugin implementation.

        This method can perform any actions required to load the plugin beyond simply returning
        the implementation.

        @return An object appropriate for the plugin type, which normally would be a class object.
        """
        pass

    @property
    def options(self) -> List[OptionInfo]:
        """@brief A list of options added by the plugin.
        @return List of @ref pyocd.core.options.OptionInfo "OptionInfo" objects.
        """
        return []

    @property
    def version(self) -> str:
        """@brief Current release version of the plugin.

        The default implementation returns pyOCD's version.

        @return String with the plugin's version, such as '2.13.4'.
        """
        return pyocd_version

    @property
    def name(self) -> str:
        """@brief Name of the plugin."""
        raise NotImplementedError()

    @property
    def description(self) -> str:
        """@brief Short description of the plugin."""
        return ""

def load_plugin_classes_of_type(plugin_group: str, plugin_dict: Dict[str, Any], base_class: type) -> None:
    """@brief Helper method to load plugins.

    Plugins are expected to return an implementation class from their Plugin.load() method. This
    class must be derived from `base_class`.

    @param plugin_group String of the plugin group, e.g. 'pyocd.probe'.
    @param plugin_dict Dictionary to fill with loaded plugin classes.
    @param base_class The required superclass for plugin implementation classes.
    """
    for entry_point in entry_points(group=plugin_group):
        # Instantiate the plugin class.
        plugin = entry_point.load()()
        if not isinstance(plugin, Plugin):
            LOG.warning("Plugin '%s' of type '%s' has an invalid plugin object",
                    entry_point.name, plugin_group)
            continue

        # Ask the plugin whether it should be loaded.
        if plugin.should_load():
            # Load the plugin and stuff the implementation class it gives
            impl_class = plugin.load()
            if not issubclass(impl_class, base_class):
                LOG.warning("Plugin '%s' of type '%s' returned an unexpected implementation class",
                        plugin.name, plugin_group)
                continue
            plugin_dict[plugin.name] = impl_class

            # Add any plugin options.
            add_option_set(plugin.options)

